from dataclasses import dataclass, field
from typing import Optional, Literal
from .infrastructure.types import DynamoDBKeyT, DynamoDBStreamT, DynamoDBBillingModeT
from fluxional.core.tools import LookupKey

_OTEL_ENVS = [
    "OTEL_EXPORTER_OTLP_ENDPOINT",
    "OTEL_EXPORTER_OTLP_HEADERS",
    "OTEL_SERVICE_NAME",
    "AWS_LAMBDA_EXEC_WRAPPER",
]


def _default_dynamodb_pk() -> DynamoDBKeyT:
    return {"key_name": "pk", "key_type": "string"}


def _default_dynamodb_sk() -> DynamoDBKeyT:
    return {"key_name": "sk", "key_type": "string"}


def _default_build_environment() -> list[str]:
    return [LookupKey.handler_context] + _OTEL_ENVS


@dataclass
class SystemSettings:
    # IDs
    default_dynamodb_id: str = field(default="fluxional_dynamodb")
    default_websocket_lambda_id: str = field(default="fluxional_websocket_lambda")
    default_api_lambda_id: str = field(default="fluxional_api_lambda")
    default_api_gateway_id: str = field(default="fluxional_api_gateway")
    default_websocket_gateway_id: str = field(default="fluxional_websocket_gateway")
    default_storage_bucket_id: str = field(default="fluxional_storage_bucket")
    default_storage_lambda_id: str = field(default="fluxional_storage_lambda")
    default_rate_schedule_task_lambda_id: str = field(
        default="fluxional_rate_schedule_task_lambda"
    )
    default_cron_schedule_task_lambda_id: str = field(
        default="fluxional_cron_schedule_task_lambda"
    )
    default_event_lambda_id: str = field(default="fluxional_event_lambda")
    default_event_queue_id: str = field(default="fluxional_event_queue")
    # Build
    build_environment: list[str] = field(default_factory=_default_build_environment)
    # Lambda Handler Override
    lambda_handler: Optional[str] = field(default=None)
    # Container CMD Override
    container_command: Optional[str] = field(default=None)
    # OTel
    aws_lambda_exec_wrapper: str = field(default="/opt/otel-instrument")


@dataclass
class PermissionSettings:
    # Database
    allow_api_to_write_to_db: bool = field(default=False)
    allow_api_to_read_from_db: bool = field(default=False)
    allow_websockets_to_read_from_db: bool = field(default=False)
    allow_websockets_to_write_to_db: bool = field(default=False)
    allow_storage_to_write_to_db: bool = field(default=False)
    allow_storage_to_read_from_db: bool = field(default=False)
    allow_events_to_write_to_db: bool = field(default=False)
    allow_events_to_read_from_db: bool = field(default=False)
    allow_tasks_to_write_to_db: bool = field(default=False)
    allow_tasks_to_read_from_db: bool = field(default=False)
    # Storage
    allow_api_to_write_to_storage: bool = field(default=False)
    allow_api_to_read_from_storage: bool = field(default=False)
    allow_api_to_delete_from_storage: bool = field(default=False)
    allow_websockets_to_write_to_storage: bool = field(default=False)
    allow_websockets_to_read_from_storage: bool = field(default=False)
    allow_websockets_to_delete_from_storage: bool = field(default=False)
    allow_storage_to_write_to_storage: bool = field(default=False)
    allow_storage_to_read_from_storage: bool = field(default=False)
    allow_storage_to_delete_from_storage: bool = field(default=False)
    allow_events_to_write_to_storage: bool = field(default=False)
    allow_events_to_read_from_storage: bool = field(default=False)
    allow_events_to_delete_from_storage: bool = field(default=False)
    allow_tasks_to_write_to_storage: bool = field(default=False)
    allow_tasks_to_read_from_storage: bool = field(default=False)
    allow_tasks_to_delete_from_storage: bool = field(default=False)

    # Events
    allow_api_to_run_events: bool = field(default=False)
    allow_websockets_to_run_events: bool = field(default=False)
    allow_storage_to_run_events: bool = field(default=False)
    allow_events_to_run_events: bool = field(default=False)
    allow_tasks_to_run_events: bool = field(default=False)


@dataclass
class DynamoDBSettings:
    default_primary_key: DynamoDBKeyT = field(default_factory=_default_dynamodb_pk)
    default_secondary_key: DynamoDBKeyT = field(default_factory=_default_dynamodb_sk)
    default_remove_on_delete: bool = field(default=True)
    default_billing_mode: DynamoDBBillingModeT = field(default="pay_per_request")
    default_stream: DynamoDBStreamT = field(default="new_image")


@dataclass
class CredentialsSettings:
    aws_account_id: Optional[str] = field(default=None)
    aws_region: Optional[str] = field(default=None)
    aws_access_key_id: Optional[str] = field(default=None)
    aws_secret_access_key: Optional[str] = field(default=None)


@dataclass
class LambdaSettings:
    directory: str = field(default="./")
    dockerfile: str = field(default="Dockerfile")
    memory_size: Literal[128, 512, 1024, 1536] = field(default=128)
    timeout: int = field(default=30)
    description: str = field(default="")


@dataclass
class ApiGatewaySettings:
    stage_name: str = field(default="prod")
    deploy: bool = field(default=True)
    description: str = field(default="fluxional_api_gateway")
    endpoint_type: Literal["regional", "edge"] = field(default="regional")


@dataclass
class WebsocketSettings:
    stage_name: str = field(default="prod")
    auto_deploy: bool = field(default=True)


@dataclass
class BuildSettings:
    dependencies: list[str] = field(default_factory=list)
    environment: dict[str, str] = field(default_factory=dict)
    requirements_file: Optional[str] = field(default="requirements.txt")
    build_path: Optional[str] = field(default=None)
    api_lambda: LambdaSettings = field(default_factory=LambdaSettings)
    storage_lambda: LambdaSettings = field(default_factory=LambdaSettings)
    event_lambda: LambdaSettings = field(default_factory=LambdaSettings)
    api_gateway: ApiGatewaySettings = field(default_factory=ApiGatewaySettings)
    websocket: WebsocketSettings = field(default_factory=WebsocketSettings)
    dynamodb: DynamoDBSettings = field(default_factory=DynamoDBSettings)
    scheduled_task_lambda: LambdaSettings = field(default_factory=LambdaSettings)


@dataclass
class OpenTelemetrySettings:
    enable: bool = field(default=False)
    exporter_otlp_endpoint: str = field(default="https://")
    exporter_otlp_headers: str = field(default="")
    service_name: str = field(default="")


@dataclass
class FluxionalTelemetrySettings:
    enable: bool = field(default=False)


@dataclass
class MonitoringSettings:
    otel: OpenTelemetrySettings = field(default_factory=OpenTelemetrySettings)
    telemetry: FluxionalTelemetrySettings = field(
        default_factory=FluxionalTelemetrySettings
    )


@dataclass
class DevelopmentSettings:
    enable_local: bool = field(default=False)


@dataclass
class StorageSettings:
    enable: bool = field(default=False)
    remove_on_delete: bool = field(default=True)


@dataclass
class Settings:
    stack_name: str = ""
    # Credentials
    credentials: CredentialsSettings = field(default_factory=CredentialsSettings)

    # Build
    build: BuildSettings = field(default_factory=BuildSettings)

    # Permissions
    permissions: PermissionSettings = field(default_factory=PermissionSettings)

    # System settings
    system: SystemSettings = field(default_factory=SystemSettings)

    # Dev
    development: DevelopmentSettings = field(default_factory=DevelopmentSettings)

    # Monitoring
    monitoring: MonitoringSettings = field(default_factory=MonitoringSettings)

    # Storage
    storage: StorageSettings = field(default_factory=StorageSettings)

    def configure(
        self,
        *,
        aws_account_id: str | None = None,
        aws_region: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        dependencies: list[str] = [],
        environment: dict[str, str] = {},
        requirements_file: str | None = "requirements.txt",
    ) -> None:
        self.credentials.aws_access_key_id = aws_access_key_id
        self.credentials.aws_secret_access_key = aws_secret_access_key
        self.credentials.aws_region = aws_region
        self.credentials.aws_account_id = aws_account_id
        self.build.dependencies = dependencies
        self.build.environment = environment
        self.build.requirements_file = requirements_file



from fluxional.types import HandlerFunctionT, AsyncHandlerFunctionT
from .infrastructure.types import RateDurationUnitT
from .handlers import Handlers
from .app import App
import functools


class Websocket:
    def __init__(self, app: App, handlers: Handlers) -> None:
        self._app = app
        self._handlers = handlers

    def on_connect(self, handler: HandlerFunctionT | AsyncHandlerFunctionT):
        # Add the connect route
        self._app.websocket.add_connect_route()
        # Add route to handlers
        self._handlers.add_websocket_route_handler("$connect", handler)

        return handler

    def on_disconnect(self, handler: HandlerFunctionT | AsyncHandlerFunctionT):
        # Add the disconnect route
        self._app.websocket.add_disconnect_route()
        # Add route to handlers
        self._handlers.add_websocket_route_handler("$disconnect", handler)

        return handler

    def default(self, handler: HandlerFunctionT | AsyncHandlerFunctionT):
        # Add the default route
        self._app.websocket.add_default_route()
        # Add route to handlers
        self._handlers.add_websocket_route_handler("$default", handler)

        return handler

    def on(
        self,
        route: str,
        handler: HandlerFunctionT | AsyncHandlerFunctionT | None = None,
    ):
        if handler is None:

            def decorator(handler: HandlerFunctionT | AsyncHandlerFunctionT):
                # Add the route
                self._app.websocket.add_route(route)
                # Add route to handlers
                self._handlers.add_websocket_route_handler(route, handler)

                @functools.wraps(handler)
                def wrapper(*args, **kwargs):
                    return handler(*args, **kwargs)

                return wrapper

            return decorator

        else:
            # Add the route
            self._app.websocket.add_route(route)
            # Add route to handlers
            self._handlers.add_websocket_route_handler(route, handler)
            return handler


class Storage:
    def __init__(self, app: App, handlers: Handlers) -> None:
        self._app = app
        self._handlers = handlers

    def on_upload(self, handler: HandlerFunctionT | AsyncHandlerFunctionT):
        self._app.storage.active = True
        self._handlers.add_storage_handler("create", handler)
        return handler

    def on_delete(self, handler: HandlerFunctionT | AsyncHandlerFunctionT):
        self._app.storage.active = True
        self._handlers.add_storage_handler("delete", handler)
        return handler


class Run:
    def __init__(self, app: App, handlers: Handlers) -> None:
        self._app = app
        self._handlers = handlers

    def every(self, value: int | float, unit: RateDurationUnitT):
        def decorator(handler: HandlerFunctionT | AsyncHandlerFunctionT):
            function_name = handler.__name__
            # add to app
            self._app.schedule.rate_schedule.append(
                {
                    "schedule_name": function_name,
                    "value": value,
                    "unit": unit,
                }
            )

            # Add the handler to handlers
            self._handlers._rate_schedule_handlers[function_name] = handler

            return handler

        return decorator

    def on(
        self,
        day: str | None = None,
        hour: str | None = None,
        minute: str | None = None,
        month: str | None = None,
        week_day: str | None = None,
        year: str | None = None,
    ):
        def decorator(handler: HandlerFunctionT | AsyncHandlerFunctionT):
            function_name = handler.__name__
            # add to app
            self._app.schedule.cron_schedule.append(
                {
                    "schedule_name": function_name,
                    "day": day,
                    "hour": hour,
                    "minute": minute,
                    "month": month,
                    "week_day": week_day,
                    "year": year,
                }
            )

            # Add the handler to handlers
            self._handlers._cron_schedule_handlers[function_name] = handler

            return handler

        return decorator


class Event:
    def __init__(self, app: App, handlers: Handlers) -> None:
        self._app = app
        self._handlers = handlers

    def event(
        self,
        handler: HandlerFunctionT | AsyncHandlerFunctionT,
    ):
        function_name = handler.__name__
        self._app.event.active = True
        self._handlers._sqs_handlers[function_name] = handler
        return handler


class LogicMixin:
    def __init__(
        self,
        *,
        websocket: Websocket,
        handlers: Handlers,
        app: App,
        storage: Storage,
        run: Run,
        event: Event,
    ) -> None:
        self._websocket = websocket
        self._handlers = handlers
        self._app = app
        self._storage = storage
        self._run = run
        self._event = event

    @property
    def websocket(self) -> Websocket:
        return self._websocket

    @property
    def storage(self) -> Storage:
        return self._storage

    @property
    def run(self) -> Run:
        return self._run

    def event(
        self,
        handler: HandlerFunctionT | AsyncHandlerFunctionT,
    ):
        self._event.event(handler)
        return handler

    def api(self, handler: HandlerFunctionT | AsyncHandlerFunctionT):
        self.add_api(handler)
        return handler

    def add_api(
        self,
        handler: HandlerFunctionT | AsyncHandlerFunctionT,
    ) -> None:
        """
        Add an API handler to the application
        """
        self._handlers.add_api_handler(handler)
        self._app.set_api()



from typing import TypedDict, Literal

# RESOURCES
ServerlessResourcesT = Literal[
    "lambda_function",
    "api_gateway",
    "ws_gateway",
    "dynamodb",
    "sns_topic",
    "s3_bucket",
    "rate_schedule",
    "cron_schedule",
    "sqs_queue",
]

# DYNAMODB
DynamoDBBillingModeT = Literal["pay_per_request"]
DynamoDBStreamT = Literal["new_image", "old_image", "new_and_old_images"]
DynamoDBAttributeTypeT = Literal["string", "number", "binary"]
RateDurationUnitT = Literal["days", "hours", "minutes", "seconds", "milliseconds"]


class DynamoDBKeyT(TypedDict):
    key_name: str
    key_type: DynamoDBAttributeTypeT


class DynamoDBLsiT(TypedDict):
    index_name: str
    sort_key: DynamoDBKeyT


class DynamoDBGsiT(TypedDict):
    index_name: str
    partition_key: DynamoDBKeyT
    sort_key: DynamoDBKeyT



class NoHandlerFound(Exception):
    """
    Exception raised when no handler is found for the event
    """


class MissingStackResource(Exception):
    """
    Exception raised when a stack resource is missing
    """


class FailedToSyncProject(Exception):
    """
    Exception raised when a project fails to sync
    """



from awscrt import auth  # type: ignore
from awsiot import mqtt_connection_builder, mqtt  # type: ignore
import boto3  # type: ignore
from uuid import uuid4
from typing import Callable
import json


class AWSIotConnection:
    def __init__(
        self,
        *,
        endpoint: str | None = None,
        client_id: str | None = None,
        region: str | None = "us-east-1",
        clean_session: bool = False,
        keep_alive_secs: int = 30,
    ):
        self._endpoint = endpoint or self.default_iot_endpoint()
        self._client_id = client_id or self.default_client_id()
        self._region = region
        self._clean_session = clean_session
        self._keep_alive_secs = keep_alive_secs

    def credentials_provider(self) -> auth.AwsCredentialsProvider:
        return auth.AwsCredentialsProvider.new_default_chain()

    @staticmethod
    def default_client_id() -> str:
        return "test-" + str(uuid4())

    @staticmethod
    def default_iot_endpoint() -> str:
        client = boto3.client("iot")
        response = client.describe_endpoint(endpointType="iot:Data-ATS")
        return response["endpointAddress"]

    def get_connection(self):
        return mqtt_connection_builder.websockets_with_default_aws_signing(
            endpoint=self._endpoint,
            region=self._region,
            credentials_provider=self.credentials_provider(),
            on_connection_interrupted=self.on_connection_interrupted,
            on_connection_resumed=self.on_connection_resumed,
            client_id=self._client_id,
            clean_session=self._clean_session,
            keep_alive_secs=self._keep_alive_secs,
        )

    def on_connection_interrupted(self, connection, error, **kwargs):
        pass

    def on_connection_resumed(self, connection, return_code, session_present, **kwargs):
        pass


class DevClient(AWSIotConnection):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._connection: mqtt.Connection = self.get_connection()

    def connect(self) -> None:
        connect_future = self._connection.connect()
        connect_future.result()

    def subscribe(
        self,
        topic: str,
        on_msg_received: Callable,
        qos: mqtt.QoS = mqtt.QoS.AT_LEAST_ONCE,
    ) -> None:
        subscribe_future, _ = self._connection.subscribe(
            topic=topic, qos=qos, callback=on_msg_received
        )
        subscribe_future.result()

    def unsubscribe(self, topic: str) -> None:
        self._connection.unsubscribe(topic=topic)

    def disconnect(self) -> None:
        disconnect_future = self._connection.disconnect()
        disconnect_future.result()

    def publish(
        self,
        topic,
        payload: dict,
        qos: mqtt.QoS = mqtt.QoS.AT_LEAST_ONCE,
        retain: bool = False,
    ):
        future, *rest = self._connection.publish(
            topic=topic, payload=json.dumps(payload), qos=qos, retain=retain
        )

        return future



# Required mounted volumes to run docker inside docker
DID_VOLUME = {"/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}}
# Base image used to run docker inside docker
BASE_IMAGE = "fluxionality/cdk_deploy:latest"
# Base workdir
BASE_WORKDIR = "/app"
# CDK DEPLOYMENT COMMAND
CDK_DEPLOY_CMD = "cdk deploy --app 'python3 {app} --synth' --require-approval never"
CDK_DESTROY_CMD = "cdk destroy --app 'python3 {app} --synth' --force"
# AWS LAMBDA BASE IMAGE
AWS_LAMBDA_PYTHON_3_10_IMAGE = "public.ecr.aws/lambda/python:3.10"
AWS_LAMBDA_PYTHON_3_11_IMAGE = "public.ecr.aws/lambda/python:3.11"
AWS_LAMBDA_PYTHON_3_12_IMAGE = "public.ecr.aws/lambda/python:3.12"
RUNTIMES = {
    "3.10": AWS_LAMBDA_PYTHON_3_10_IMAGE,
    "3.11": AWS_LAMBDA_PYTHON_3_11_IMAGE,
    "3.12": AWS_LAMBDA_PYTHON_3_12_IMAGE,
}

# OTEL
OTEL_WRAPPER_FILE = "https://tjaws.s3.amazonaws.com/otel_wrapper.py"
OTEL_INST_FILE = "https://tjaws.s3.amazonaws.com/otel-instrument"
OTEL_REQ_FILE = "https://tjaws.s3.amazonaws.com/requirements_otel.txt"

# Fluxional Telemetry Extensions
FTE_EXTENSION = (
    "https://tjaws.s3.amazonaws.com/python-example-telemetry-api-extension.zip"
)



import json
from rich import print as rp
from rich.console import Console
from .client import DevClient
from uuid import uuid4
from typing import Any, Callable
from .runner import run, build_runner_image
import queue as q
import threading
from dataclasses import dataclass
from fluxional.core.tools import LookupKey


def run_container(
    *,
    stack_name: str,
    payload: Any,
    subscriber_id: str,
    runner_function: Callable = run,
):
    try:
        payload = json.loads(payload.decode("utf-8"))
        event_id = payload["event_id"]
        event = payload["event"]
        execution_context = payload["execution_context"]
        del execution_context[LookupKey.handler_context]

    except KeyError as e:
        raise ValueError(f"{e} is required in payload")

    container = runner_function(
        stack_name,
        environment={
            "EVENT_ID": event_id,
            "EVENT": json.dumps(event),
            "SUBSCRIBER_ID": subscriber_id,
            "PYTHONUNBUFFERED": "1",
            **execution_context,
        },
    )
    return container


def worker(
    queue: q.Queue, worker_id, color: str, container_provider: Callable = run_container
):
    console = Console(style=color)
    while True:
        item = queue.get()
        if item is None:
            break

        console.print(f"Worker: {worker_id} working on item queue")
        container_provider(
            stack_name=item["stack_name"],
            payload=item["kwargs"]["payload"],
            subscriber_id=item["subscriber_id"],
        )

        queue.task_done()


@dataclass
class Workers:
    num_workers: int = 5
    colors = ["red", "green", "blue", "yellow", "magenta"]
    queue: q.Queue[Any] = q.Queue[Any]()
    thread: threading.Thread | None = None
    worker: Callable = worker

    def start(self):
        assert len(self.colors) >= self.num_workers, "Not enough colors for workers"
        for i in range(self.num_workers):
            t = threading.Thread(
                target=self.worker,
                args=(self.queue, i, self.colors[i % len(self.colors)]),
            )
            t.start()
            self.thread = t

    def stop(self):
        self.queue.join()
        for _ in range(self.num_workers):
            self.queue.put(None)
        if self.thread:
            self.thread.join()

    def add_to_queue(self, item: dict):
        self.queue.put(item)


def run_forever(
    workers: Workers, client: DevClient, loop_function: Callable | None = None
):
    try:
        while True:
            if loop_function:
                loop_function()

    except KeyboardInterrupt:
        pass

    finally:
        workers.stop()
        rp("Disconnecting from server")
        client.disconnect()


def run_dev(
    stack_name: str,
    *,
    handler: str,
    dev_client: type[DevClient] = DevClient,
    build_function: Callable = build_runner_image,
    workers_provider: type[Workers] = Workers,
    run_forever: Callable = run_forever,
):
    subscriber_id = str(uuid4())

    rp("\n[bold blue]Starting Development[/bold blue]")
    client: DevClient = dev_client()

    client.connect()

    rp("[blue]Building Dev Environment[/blue]")
    build_function(stack_name, handler)

    rp("[green]Starting workers[/green]")
    workers = workers_provider()
    workers.start()

    client.subscribe(
        f"fluxional/events/{stack_name}",
        lambda **rest: workers.add_to_queue(
            {"stack_name": stack_name, "subscriber_id": subscriber_id, "kwargs": rest}
        ),
    )

    rp("[green]Connected To Server - Ready to receive events[/green]")

    run_forever(workers, client)



from fluxional.core import (
    Fluxional,
    Environment,
    Extender,
    Settings,
    ApiEvent,
    LambdaContext,
    WsEvent,
    Event,
    Websocket,
    TaskEvent,
    StorageEvent,
)

__all__ = [
    "Fluxional",
    "Environment",
    "Extender",
    "Settings",
    "ApiEvent",
    "LambdaContext",
    "WsEvent",
    "Event",
    "Websocket",
    "TaskEvent",
    "StorageEvent",
]


__version__ = "0.1.6"


from fluxional.deployment.engine import Engine
from fluxional.deployment.constants import AWS_LAMBDA_PYTHON_3_10_IMAGE
import os


class FunctionEngine(Engine):
    def setup_credentials(self):
        pass


def build_launcher(handler: str) -> str:
    file, module = handler.split(".")

    code = [f"from {file} import {module}"]
    code.append("from fluxional.dev.client import DevClient")
    code.append("import os")
    code.append("import json")
    code.append(r"client = DevClient()")
    code.append(r"client.connect()")
    code.append(r"event = json.loads(os.environ.get('EVENT'))")
    code.append(r"event_id = os.environ.get('EVENT_ID')")
    code.append(r"print('EVENT_ID: ', event_id)")
    code.append(r"subscriber_id = os.environ.get('SUBSCRIBER_ID')")
    code.append(r"print('Acknowledging event...')")
    code.append(
        r"x = client.publish(f'fluxional/acknowledge/{event_id}', {'subscriber_id': subscriber_id})"
    )
    code.append(r"x.result(timeout=5)")
    code.append(r"print('Event acknowledged')")
    code.append("try:")
    code.append(rf"    result = {module}(event, None)")
    code.append(r"except Exception as e:")
    code.append(r"    print('ERROR:', e)")
    code.append(r"    result = {'statusCode': 500, 'body': 'Internal Server Error'}")
    code.append(r"print(result)")
    code.append(r"print('Sending response...')")
    code.append(
        r"x = client.publish(f'fluxional/response/{event_id}', {'subscriber_id': subscriber_id, 'response': result})"
    )
    code.append(r"x.result(timeout=5)")
    code.append(r"print('Response sent')")
    code.append(r"client.disconnect()")

    return "\\n".join(code)


def build_runner_image(
    stack_name: str,
    handler: str,
    *,
    requirements_file: str | None = "requirements.txt",
    build_path: str = ".",
    engine_provider: type[FunctionEngine] = FunctionEngine,
):
    dockerfile = f"FROM {AWS_LAMBDA_PYTHON_3_10_IMAGE}"

    dockerfile += "\nWORKDIR /"

    if requirements_file:
        dockerfile += f"\nCOPY {requirements_file} ."
        dockerfile += f"\nRUN pip install -r {requirements_file}"

    # Remove the default entrypoint
    dockerfile += "\nENTRYPOINT []"
    # Create the launcher file
    launcher = build_launcher(handler)
    dockerfile += f'\n RUN echo -e "{launcher}" >> launcher.py'
    dockerfile += "\nENV PYTHONPATH /app"
    stack_name = stack_name.lower()
    tag = f"{stack_name}_local_runner"
    engine = engine_provider(tag=tag, build_path=build_path, remove_container=False)
    engine.build_image(dockerfile, show_logs=True)


def run(
    stack_name: str,
    command: str = "python3 launcher.py",
    build_path: str = ".",
    environment: dict = {},
    engine_provider: type[FunctionEngine] = FunctionEngine,
):
    # Mout the current directory to /app
    mount_path = os.path.join(os.getcwd(), build_path)
    mount_volume = {mount_path: {"bind": "/app", "mode": "ro"}}

    engine = engine_provider(
        tag=f"{stack_name.lower()}_local_runner",
        build_path=build_path,
        remove_container=False,
    )

    if engine.image_exists():
        engine.run_container(
            command=command,
            detach=True,
            show_logs=True,
            volumes=mount_volume,
            environment=environment,
            network_mode="host",
        )

    return



from typing import Literal


def is_http_event(event, context) -> bool:
    if "httpMethod" in event:
        return True

    return False


def is_sns_event(event, context) -> bool:
    if "Records" in event:
        for record in event["Records"]:
            if "EventSource" in record and record["EventSource"] == "aws:sns":
                return True

    return False


def is_fluxional_event(event, context) -> bool:
    if "fluxional_event" in event:
        return True

    return False


def is_websocket_event(event, context) -> str | bool:
    """Check if websocket event and return the route to it"""
    if "requestContext" in event:
        if "routeKey" in event["requestContext"]:
            return event["requestContext"]["routeKey"]

    return False


def is_eventbridge_scheduled_event(event, context) -> bool:
    if "source" in event and event["source"] == "aws.events":
        if "detail-type" in event and event["detail-type"] == "Scheduled Event":
            return True

    return False


def is_rate_schedule_event(event, context) -> bool:
    if "schedule_type" in event and event["schedule_type"] == "RateSchedule":
        return True

    return False


def is_cron_schedule_event(event, context) -> bool:
    if "schedule_type" in event and event["schedule_type"] == "CronSchedule":
        return True

    return False


def is_dev_context(event, context) -> bool:
    """Check if the current context is a realtime lambda"""
    from .tools import Environment

    env = Environment()

    if env.fluxional_handler_context == "development":
        # Only support http events for now
        if is_http_event(event, context):
            return True

    return False


S3_BUCKET_CREATE_EVENTS = [
    "ObjectCreated:*",
    "ObjectCreated:Put",
    "ObjectCreated:Post",
    "ObjectCreated:Copy",
    "ObjectCreated:CompleteMultipartUpload",
]


S3_BUCKET_DELETE_EVENTS = [
    "ObjectRemoved:*",
    "ObjectRemoved:Delete",
    "ObjectRemoved:DeleteMarkerCreated",
]

S3_ACTIONS = Literal["create", "delete"]


def is_s3_bucket_create_event(event, context) -> bool:
    if "Records" in event:
        for record in event["Records"]:
            if "eventSource" in record and record["eventSource"] == "aws:s3":
                if record["eventName"] in S3_BUCKET_CREATE_EVENTS:
                    return True

    return False


def is_s3_bucket_delete_event(event, context) -> bool:
    if "Records" in event:
        for record in event["Records"]:
            if "eventSource" in record and record["eventSource"] == "aws:s3":
                if record["eventName"] in S3_BUCKET_DELETE_EVENTS:
                    return True

    return False


def is_sqs_event(event, context) -> bool:
    if "Records" in event:
        for record in event["Records"]:
            if "eventSource" in record and record["eventSource"] == "aws:sqs":
                return True

    return False



from .handlers import Handlers
from typing import Any
from .settings import Settings
from .app import App
import os
from .infrastructure.types import DynamoDBKeyT, DynamoDBLsiT, DynamoDBGsiT
from .logic import Websocket, LogicMixin, Storage, Run, Event

cwd = os.getcwd()


class Extender(LogicMixin):
    def __init__(self) -> None:
        self._settings = Settings("Extender")
        self._app = App(settings=self._settings)
        self._handlers: Handlers = Handlers(settings=self._settings, app=self._app)
        self._websocket = Websocket(app=self._app, handlers=self._handlers)
        self._storage = Storage(app=self._app, handlers=self._handlers)
        self._run = Run(app=self._app, handlers=self._handlers)
        self._event = Event(app=self._app, handlers=self._handlers)
        super().__init__(
            websocket=self._websocket,
            handlers=self._handlers,
            app=self._app,
            storage=self._storage,
            run=self._run,
            event=self._event,
        )


class Fluxional(LogicMixin):
    def __init__(
        self,
        stack_name: str,
    ) -> None:
        self._stack_name = stack_name
        self._settings = Settings(stack_name=stack_name)
        self._app = App(settings=self._settings)
        self._handlers: Handlers = Handlers(settings=self._settings, app=self._app)
        self._websocket = Websocket(app=self._app, handlers=self._handlers)
        self._storage = Storage(app=self._app, handlers=self._handlers)
        self._run = Run(app=self._app, handlers=self._handlers)
        self._event = Event(app=self._app, handlers=self._handlers)
        super().__init__(
            websocket=self._websocket,
            handlers=self._handlers,
            app=self._app,
            storage=self._storage,
            run=self._run,
            event=self._event,
        )

    @property
    def settings(self) -> Settings:
        return self._settings

    def add_dynamodb(
        self,
        *,
        partition_key: DynamoDBKeyT | None = None,
        sort_key: DynamoDBKeyT | None = None,
        remove_on_delete: bool = True,
        local_secondary_indexes: list[DynamoDBLsiT] | None = None,
        global_secondary_indexes: list[DynamoDBGsiT] | None = None,
    ) -> None:
        return self._app.add_dynamodb(
            partition_key=partition_key,
            sort_key=sort_key,
            remove_on_delete=remove_on_delete,
            local_secondary_indexes=local_secondary_indexes,
            global_secondary_indexes=global_secondary_indexes,
        )

    def configure(
        self,
        *,
        aws_account_id: str | None = None,
        aws_region: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        dependencies: list[str] = [],
        environment: dict[str, str] = {},
        requirements_file: str | None = "requirements.txt",
    ) -> None:
        """
        Configure the settings for the project
        """
        return self._settings.configure(
            aws_account_id=aws_account_id,
            aws_region=aws_region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            dependencies=dependencies,
            environment=environment,
            requirements_file=requirements_file,
        )

    def synth(self):
        """
        Generate the infrastructure Cloudformation template
        using cdk synth
        """
        from .infrastructure.base import Infrastructure
        from .infrastructure.resources import InfrastructureT, InfraSettings

        # Resolve environment variables by attempting to find
        # it in the os if it is not a static value
        environment = {
            k: os.environ.get(k, v)
            for k, v in self._settings.build.environment.items()
            # @REFRACTOR
            # This needs to be refractored k cannot be none here
            # If an environment variable of None is passed it will fail
            if (k, v) != (None, None)
        }
        # Resolve build environment
        build_environment = {
            k: os.environ.get(k)
            for k in self._settings.system.build_environment
            if os.environ.get(k)
        }

        infra = Infrastructure(
            infrastructure=InfrastructureT(
                settings=InfraSettings(
                    stack_name=self._stack_name,
                    aws_account_id=self._settings.credentials.aws_account_id,
                    aws_region=self._settings.credentials.aws_region,
                    environment=environment | build_environment,
                ),
                resources=self._app.build_resources(),
            )
        )

        return infra.stack().app.synth()

    def handler(self) -> Any:
        """
        Main entrypoint for all events
        """

        cli = self._handlers.synth_handler()
        if cli is not None:
            return self.synth()

        return self._handlers.handler()

    def register(self, extender: Extender) -> None:

        def _append_only(handlers: dict, extender_handlers: dict):
            for key, value in extender_handlers.items():
                if key not in handlers:
                    handlers[key] = value

        # At this point we will just override for the api
        if extender._app.api.active:
            self._app.api = extender._app.api
            self._handlers._http_handlers = extender._handlers._http_handlers

        if extender.websocket._app.websocket.routes:
            self._app.websocket = extender._app.websocket
            _append_only(
                self._handlers._websocket_handlers,
                extender._handlers._websocket_handlers,
            )

        if self._settings.storage.enable or extender._app.storage.active:
            self._app.storage = extender._app.storage
            _append_only(
                self._handlers._storage_handlers,
                extender._handlers._storage_handlers,
            )

        if extender._app.schedule.rate_schedule or extender._app.schedule.cron_schedule:
            self._app.schedule = extender._app.schedule
            _append_only(
                self._handlers._rate_schedule_handlers,
                extender._handlers._rate_schedule_handlers,
            )
            _append_only(
                self._handlers._cron_schedule_handlers,
                extender._handlers._cron_schedule_handlers,
            )

        if extender._app.event.active:
            self._app.event = extender._app.event
            _append_only(
                self._handlers._sqs_handlers,
                extender._handlers._sqs_handlers,
            )

    def set_settings(self, settings: Settings) -> None:
        settings.stack_name = self._stack_name
        self._settings = settings
        self._app.settings = settings
        self._handlers._settings = settings



from typing import Callable, Any, TypeGuard, Awaitable
import inspect

# Represents a lambda function style handler Accepts (event, context)
HandlerFunctionT = Callable[[Any, Any], Any]
AsyncHandlerFunctionT = Callable[[Any, Any], Awaitable[Any]]


class AsyncTypeGuard:
    @staticmethod
    def is_async(
        handler: HandlerFunctionT | AsyncHandlerFunctionT,
    ) -> TypeGuard[AsyncHandlerFunctionT]:
        return inspect.iscoroutinefunction(handler)



from .exceptions import (
    FailedImageBuild,
    FailedContainerBuild,
    FailedToRunContainer,
)
from .constants import (
    DID_VOLUME,
    BASE_WORKDIR,
    BASE_IMAGE,
    CDK_DEPLOY_CMD,
    CDK_DESTROY_CMD,
    AWS_LAMBDA_PYTHON_3_10_IMAGE,
    RUNTIMES,
    OTEL_WRAPPER_FILE,
    OTEL_INST_FILE,
    OTEL_REQ_FILE,
    FTE_EXTENSION,
)
import docker  # type: ignore
import os
import json
from dotenv import load_dotenv
from fluxional.utils import (
    default_aws_account_id,
    default_aws_region,
    default_aws_secret_access_key,
    default_aws_access_key_id,
)
from rich.console import Console
from rich import print as rp
from typing import Any, Literal
from fluxional import __version__
import re


load_dotenv()

# Fix that will avoid using fileobject which does not allow
# Context - IE: Imposible to copy files from host to container
docker.api.build.process_dockerfile = lambda file, _: (
    "Dockerfile",
    file,
)

console = Console()

RuntimeT = Literal["3.10", "3.11", "3.12"]

_Container = type[docker.models.containers.Container]


def st(process: Any, description: str, *args, **kwargs):
    with console.status(description):
        output = process(*args, **kwargs)

    if isinstance(output, list):
        output_key_info(output)


def output_key_info(logs: list[str]):
    # Currently only extracts API Gateway Endpoints
    rp("\n[underline bold yellow]Outputs:")

    for log in logs:
        http_url_pattern = r"https://[a-zA-Z0-9./-]+"
        ws_url_pattern = r"wss://[a-zA-Z0-9./-]+"

        http_match = re.search(http_url_pattern, log)
        ws_match = re.search(ws_url_pattern, log)

        if http_match:
            api_endpoint = http_match.group()
            rp(f"\n👉 [bold yellow] API Endpoint: {api_endpoint}")

        if ws_match:
            websocket_url = ws_match.group()
            rp(f"\n👉 [bold yellow] WebSocket URL: {websocket_url}")


def track_logs(container: _Container) -> list[str]:
    logs: list[str] = []
    output_logs: list[str] = []

    for line in container.logs(stream=True):
        log = line.decode("utf-8").strip()
        if log:
            logs.append(log)
            if log == "Outputs:":
                output_logs.append(log)
            elif output_logs:
                output_logs.append(log)
            else:
                print(log)

    try:
        container_exit_code = container.wait()["StatusCode"]
        if container_exit_code != 0:
            for log in logs:
                rp(log)
            raise FailedContainerBuild("The container failed to build.")
    except docker.errors.NotFound:  # pragma: no cover
        pass

    return output_logs


class Engine:
    def __init__(
        self,
        *,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_account_id: str | None = None,
        aws_region: str | None = None,
        base_url: str = "unix://var/run/docker.sock",
        tag: str = "fluxional_deploy_image",
        build_path: str = os.getcwd(),
        remove_container: bool = False,
    ):
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_account_id = aws_account_id
        self._aws_region = aws_region
        self._tag = tag
        self._build_path = build_path
        self._remove_container = remove_container
        self._api_client = docker.APIClient(base_url=base_url)
        self._client = docker.from_env()
        self.setup_credentials()

    def setup_credentials(self):
        if not self._aws_access_key_id:
            self._aws_access_key_id = default_aws_access_key_id()

        if not self._aws_secret_access_key:
            self._aws_secret_access_key = default_aws_secret_access_key()

        if not self._aws_account_id:
            self._aws_account_id = default_aws_account_id()

        if not self._aws_region:
            self._aws_region = default_aws_region()

    def build_image(
        self,
        dockerfile: str,
        tag: str | None = None,
        show_logs: bool = False,
        rm: bool = True,
    ):
        """
        Builds the docker image given a dockerfile like strings
        """

        def display(row: Any):
            if "errorDetail" in row:
                row = json.loads(row)["errorDetail"]["message"]
                rp(f"[red]{row}")
                raise FailedImageBuild(row)

            elif '"stream"' in row:
                row = json.loads(row)
                if row["stream"] != "\n":
                    row = row["stream"].replace("\n", "")
                    if show_logs:
                        print(row)

        build = self._api_client.build(
            dockerfile=dockerfile,
            rm=rm,
            tag=tag or self._tag,
            path=self._build_path,
        )

        for line in build:
            row = line.decode("utf-8").strip()

            # Windows specific
            if "\r\n" in row:
                row = row.split("\r\n")

                for row in row:
                    display(row)

            else:
                display(row)

    def run_container(
        self,
        *,
        command: str,
        volumes: dict[str, dict[str, str]] = DID_VOLUME,
        environment: dict[str, str] = {},
        tag: str | None = None,
        detach: bool = False,
        show_logs: bool = False,
        network_mode: Literal["host", "bridge", "none"] = "bridge",
    ) -> list[str]:
        """
        Runs the built image given a command
        """

        try:
            container = self._client.containers.run(
                tag or self._tag,
                remove=True,
                auto_remove=self._remove_container,
                detach=detach,
                volumes=volumes,
                environment={
                    "AWS_ACCESS_KEY_ID": self._aws_access_key_id,
                    "AWS_SECRET_ACCESS_KEY": self._aws_secret_access_key,
                    "AWS_ACCOUNT_ID": self._aws_account_id,
                    "AWS_REGION": self._aws_region,
                }
                | environment,
                command=command,
                network_mode=network_mode,
            )

        except Exception as e:
            raise FailedToRunContainer(e)

        if not detach and show_logs:
            if isinstance(container, bytes):
                rp("\n[bold blue] Container Ouptut:\n")
                container = container.decode("utf-8")
                print(container)
                return []

        if detach and show_logs:
            return track_logs(container)

        else:
            return []

    def remove_built_image(self, tag: str | None = None, **kwargs):
        """
        Removes the built image from the docker daemon
        """

        try:
            self._client.images.remove(tag or self._tag, **kwargs)
        except docker.errors.ImageNotFound:
            pass

    def image_exists(self):
        """
        Check if the image exists
        """
        try:
            self._client.images.get(self._tag)
            return True
        except docker.errors.ImageNotFound:
            return False


class CDKEngine(Engine):
    @staticmethod
    def build_dep_copy(
        dependencies: list[str], *, copy_directory=".", work_directory=BASE_WORKDIR
    ) -> str:
        """
        Build the command COPY based on a list of dependencies
        """
        deps = [f"COPY {copy_directory}/{k} {work_directory}/{k}" for k in dependencies]

        return "\n".join(deps)

    def get_dockerfile(
        self,
        *,
        dependencies=list[str],
        requirements_file: str | None = None,
        base_image: str = BASE_IMAGE,
        work_directory: str = BASE_WORKDIR,
    ) -> str:
        """
        Create a dockerfile based on dependencies etc...
        """
        base_image = f"FROM {base_image}"
        work_dir = f"WORKDIR {work_directory}"
        dockerfile = f"{base_image}"

        if dependencies:
            files = self.build_dep_copy(dependencies, work_directory=work_directory)
            dockerfile += f"\n{files}"

        dockerfile += f"\n{work_dir}"

        if requirements_file:
            dockerfile += f"\nRUN pip install -r {requirements_file}"

        dockerfile += "\n"

        return dockerfile

    def get_otel_steps(self) -> str:
        # Download the otel files and install requirements
        dockerfile = f"\nRUN curl -sSL {OTEL_INST_FILE} -o /opt/otel-instrument"
        dockerfile += "\nRUN chmod +x /opt/otel-instrument"
        dockerfile += f"\nRUN curl -sSL {OTEL_REQ_FILE} -o requirements_otel.txt"
        dockerfile += "\nRUN pip install -r requirements_otel.txt --target /opt/python"
        dockerfile += (
            f"\nRUN curl -sSL {OTEL_WRAPPER_FILE} -o /opt/python/otel_wrapper.py"
        )
        return dockerfile

    def get_fte_steps(self) -> str:
        dockerfile = "\nRUN pip install requests"
        dockerfile += "\nRUN yum install -y unzip"
        dockerfile += f"\nRUN curl -sSL {FTE_EXTENSION} -o ./python-example-telemetry-api-extension.zip"
        dockerfile += "\nRUN unzip ./python-example-telemetry-api-extension.zip -d ./python-example-telemetry-api-extension"
        # Copy content to /opt
        dockerfile += "\nRUN cp -r ./python-example-telemetry-api-extension/python-example-telemetry-api-extension/* /opt/"
        # Remove the zip file
        dockerfile += "\nRUN rm ./python-example-telemetry-api-extension.zip"
        # Remove the folder
        dockerfile += "\nRUN rm -rf ./python-example-telemetry-api-extension"

        return dockerfile

    def get_lambda_dockefile(
        self,
        *,
        lambda_handler: str,
        base_image: str = AWS_LAMBDA_PYTHON_3_10_IMAGE,
        requirements_file: str | None = None,
        include_otel: bool = False,
        include_fte: bool = False,
    ):
        dockerfile = f"FROM {base_image}"

        dockerfile += "\nCOPY . \${LAMBDA_TASK_ROOT}"

        if requirements_file:
            dockerfile += (
                f"\nRUN pip install -r {requirements_file}"
                # Needs to be escaped once for the echo cmd
                + r' --target "\${LAMBDA_TASK_ROOT}"'
            )
        else:
            # We will always need fluxional installed
            dockerfile += (
                f"\nRUN pip install fluxional=={__version__}"
                + r' --target "\${LAMBDA_TASK_ROOT}"'
            )

        if include_otel:
            dockerfile += self.get_otel_steps()

        if include_fte:
            dockerfile += self.get_fte_steps()

        dockerfile += f'\nCMD [\\"{lambda_handler}\\"]'

        return dockerfile

    def run_clean(
        self,
        *,
        dependencies: list[str],
        requirements_file: str | None,
        handler: str,
        command: str,
        environment: dict[str, str],
        lambda_handler: str | None = None,
        show_logs: bool,
        base_image: str = AWS_LAMBDA_PYTHON_3_10_IMAGE,
        include_otel: bool = False,
        include_fte: bool = False,
    ):
        dockerfile = self.get_dockerfile(
            dependencies=dependencies, requirements_file=requirements_file
        )

        # We need to add the lambda Dockefile for cdk here
        lambda_dockerfile = self.get_lambda_dockefile(
            lambda_handler=lambda_handler if lambda_handler else handler,
            requirements_file=requirements_file,
            base_image=base_image,
            include_otel=include_otel,
            include_fte=include_fte,
        )

        for line in lambda_dockerfile.split("\n"):
            dockerfile += f'RUN echo "{line}" >> Dockerfile\n'

        st(self.build_image, "[blue]Building image", dockerfile, show_logs=show_logs)

        # Run the container
        st(
            self.run_container,
            "[blue]Deploying changes",
            command=command,
            environment=environment,
            show_logs=show_logs,
            detach=True,
        )

        # Always clean the image
        st(self.remove_built_image, "[blue]Cleaning up", force=True)

    def run(
        self,
        handler: str,
        *,
        command: str,
        dependencies: list[str],
        requirements_file: str | None,
        environment: dict[str, str] = {},
        lambda_handler: str | None = None,
        lambda_runtime: RuntimeT = "3.10",
        show_logs: bool = False,
        include_otel: bool = False,
        include_fte: bool = False,
    ):
        rp("\n\n[bold blue]Fluxional CDK Engine \U0001F680\n")

        # We need to include certain mandatory files to auto-copy
        # For the file name
        file_name = handler.split(".")[0]
        if f"{file_name}.py" not in dependencies:
            dependencies.append(f"{file_name}.py")

        # For requirements
        if requirements_file and requirements_file not in dependencies:
            dependencies.append(requirements_file)

        rp("[bold]Configurations:")
        rp(f"Show logs: {show_logs}")
        rp(f"Dependencies: {dependencies}")
        rp(f"Runtime: Python {lambda_runtime}")

        self.run_clean(
            dependencies=dependencies,
            requirements_file=requirements_file,
            handler=lambda_handler if lambda_handler else handler,
            command=command,
            environment=environment,
            lambda_handler=lambda_handler,
            show_logs=show_logs,
            base_image=RUNTIMES[lambda_runtime],
            include_otel=include_otel,
            include_fte=include_fte,
        )

        rp("\n\n[bold blue]  \U0001F680 Change Completed! \U0001F680 \n\n")

    def deploy(
        self,
        handler: str,
        *,
        command: str | None = None,
        dependencies: list[str] = [],
        requirements_file: str | None = "requirements.txt",
        environment: dict[str, str] = {},
        lambda_handler: str | None = None,
        lambda_runtime: RuntimeT = "3.10",
        show_logs: bool = False,
        include_otel: bool = False,
        include_fte: bool = False,
    ):
        file_name, _ = handler.split(".")

        if not command:
            command = CDK_DEPLOY_CMD.format(app=f"{file_name}.py")

        self.run(
            handler,
            command=command,
            dependencies=dependencies,
            requirements_file=requirements_file,
            environment=environment,
            lambda_handler=lambda_handler,
            show_logs=show_logs,
            lambda_runtime=lambda_runtime,
            include_otel=include_otel,
            include_fte=include_fte,
        )

    def destroy(
        self,
        handler: str,
        *,
        command: str | None = None,
        dependencies: list[str] = [],
        requirements_file: str | None = "requirements.txt",
        environment: dict[str, str] = {},
        lambda_handler: str | None = None,
        lambda_runtime: RuntimeT = "3.10",
        show_logs: bool = False,
    ):
        file_name, _ = handler.split(".")

        if not command:
            command = CDK_DESTROY_CMD.format(app=f"{file_name}.py")

        self.run(
            handler,
            command=command,
            dependencies=dependencies,
            requirements_file=requirements_file,
            environment=environment,
            lambda_handler=lambda_handler,
            show_logs=show_logs,
            lambda_runtime=lambda_runtime,
        )



from .resources import (
    InfraSettings,
    InfrastructureT,
    LambdaFunction,
    ApiGateway,
    WsGateway,
    AllResources,
    DynamoDB,
    ResourceTypeGuard,
    SnsTopic,
    parse_dict_to_resources,
    S3Bucket,
    RateSchedule,
    CronSchedule,
    SqsQueue,
)
from .cdk import (
    add_lambda_function_to_stack,
    add_existing_rest_api_gateway_to_stack,
    add_rest_api_gateway_to_stack,
    add_dynamodb_to_stack,
    add_sns_topic_to_stack,
    add_ws_gateway_to_stack,
    add_s3_bucket_to_stack,
    add_rate_schedule_to_stack,
    add_cron_schedule_to_stack,
    add_sqs_queue_to_stack,
)
from aws_cdk import (
    Stack as CDKStack,
    App,
    Environment,
    aws_apigateway,
    aws_iam,
    aws_lambda,
    aws_sns_subscriptions,
    aws_s3_notifications,
    aws_events_targets,
    aws_events,
    aws_sqs,
    aws_s3,
    aws_lambda_event_sources,
)
from aws_cdk.aws_apigatewayv2_integrations_alpha import WebSocketLambdaIntegration
from aws_cdk.aws_apigatewayv2 import WebSocketApi
from fluxional.exceptions import MissingStackResource
from fluxional.utils import default_aws_account_id, default_aws_region
from fluxional.core.tools import LookupKey


class Stack(CDKStack):
    def __init__(
        self,
        stack_name: str,
        *,
        aws_account_id: str,
        aws_region: str,
        environment_vars: dict = {},
    ):
        self._stack_name = stack_name
        self._aws_account_id = aws_account_id
        self._aws_region = aws_region
        self._environment_vars = environment_vars
        self._environment = Environment(account=aws_account_id, region=aws_region)
        self._app = App()
        super().__init__(self._app, self._stack_name, env=self._environment)

    @property
    def app(self) -> App:
        return self._app

    def add_sqs_queue(self, resource: SqsQueue):
        queue = add_sqs_queue_to_stack(
            stack=self,
            id=resource.id,
            queue_name=resource.queue_name,
            visibility_timeout=resource.visibility_timeout,
        )

        if resource.permissions:
            for k in resource.permissions:
                if ResourceTypeGuard.is_lambda_permission(k):

                    if k.allow_invoke:
                        func: aws_lambda.Function = getattr(self, k.resource_id)

                        source = aws_lambda_event_sources.SqsEventSource(
                            queue,
                            batch_size=1,
                        )
                        func.add_event_source(source)

    def add_cron_schedule(self, resource: CronSchedule):
        schedule = add_cron_schedule_to_stack(
            stack=self,
            id=resource.id,
            schedule_name=resource.schedule_name,
            day=resource.day,
            hour=resource.hour,
            minute=resource.minute,
            month=resource.month,
            week_day=resource.week_day,
            year=resource.year,
        )

        if resource.permissions:
            for k in resource.permissions:
                if ResourceTypeGuard.is_lambda_permission(k):
                    if k.allow_invoke:
                        schedule.add_target(
                            aws_events_targets.LambdaFunction(
                                getattr(self, k.resource_id),
                                event=aws_events.RuleTargetInput.from_object(
                                    {
                                        "schedule_type": "CronSchedule",
                                        "schedule_name": resource.schedule_name,
                                    }
                                ),
                            )
                        )

    def add_rate_schedule(self, resource: RateSchedule):
        schedule = add_rate_schedule_to_stack(
            stack=self,
            id=resource.id,
            unit=resource.unit,
            value=resource.value,
            schedule_name=resource.schedule_name,
        )

        if resource.permissions:
            for k in resource.permissions:
                if ResourceTypeGuard.is_lambda_permission(k):
                    if k.allow_invoke:
                        schedule.add_target(
                            aws_events_targets.LambdaFunction(
                                getattr(self, k.resource_id),
                                event=aws_events.RuleTargetInput.from_object(
                                    {
                                        "schedule_type": "RateSchedule",
                                        "schedule_name": resource.schedule_name,
                                    }
                                ),
                            )
                        )

    def add_sns_topic(self, resource: SnsTopic):
        sns = add_sns_topic_to_stack(
            stack=self,
            id=resource.id,
            display_name=resource.display_name,
        )

        if resource.permissions:
            for k in resource.permissions:
                if ResourceTypeGuard.is_lambda_permission(k):
                    if k.allow_invoke:
                        sns.add_subscription(
                            aws_sns_subscriptions.LambdaSubscription(
                                getattr(self, k.resource_id)
                            )
                        )

    def add_lambda_function(self, resource: LambdaFunction):
        lambda_ = add_lambda_function_to_stack(
            stack=self,
            id=resource.id,
            function_name=resource.function_name,
            directory=resource.directory,
            file=resource.dockerfile,
            memory_size=resource.memory_size,
            timeout=resource.timeout,
            description=resource.description,
        )

        # Add environment variables
        if self._environment_vars:
            for k in self._environment_vars:
                lambda_.add_environment(key=k, value=self._environment_vars[k])

                # @DEV - May be refractored to permission in the future
                # Add a policy if the fluxional_handler_context in os
                if k == LookupKey.handler_context:
                    if self._environment_vars[k] == "development":
                        lambda_.add_to_role_policy(
                            aws_iam.PolicyStatement(
                                effect=aws_iam.Effect.ALLOW,
                                actions=["iot:Connect"],
                                resources=["*"],
                            )
                        )
                        lambda_.add_to_role_policy(
                            aws_iam.PolicyStatement(
                                effect=aws_iam.Effect.ALLOW,
                                actions=["iot:Publish", "iot:Receive"],
                                resources=["arn:aws:iot:*:*:topic/fluxional*"],
                            )
                        )
                        lambda_.add_to_role_policy(
                            aws_iam.PolicyStatement(
                                effect=aws_iam.Effect.ALLOW,
                                actions=["iot:Subscribe"],
                                resources=["arn:aws:iot:*:*:topicfilter/fluxional*"],
                            )
                        )
                        lambda_.add_to_role_policy(
                            aws_iam.PolicyStatement(
                                effect=aws_iam.Effect.ALLOW,
                                actions=["iot:DescribeEndpoint"],
                                resources=["*"],
                            )
                        )

        # Add the resource id
        lambda_.add_environment(LookupKey.resource_id, resource.id)

        # Add the execution context to cloud
        lambda_.add_environment(LookupKey.execution_context, "cloud")

        if resource.permissions:
            for k in resource.permissions:
                if ResourceTypeGuard.is_dynamodb_permission(k):
                    if k.allow_read:
                        getattr(self, k.resource_id).grant_read_data(lambda_)

                    if k.allow_write:
                        getattr(self, k.resource_id).grant_write_data(lambda_)

                    lambda_.add_environment(
                        k.resource_id + "_table_name",
                        getattr(self, k.resource_id).table_name or "",
                    )

                elif ResourceTypeGuard.is_sns_permission(k):
                    if k.allow_publish:
                        sns = getattr(self, k.resource_id)

                        sns_policy = aws_iam.PolicyStatement(
                            effect=aws_iam.Effect.ALLOW,
                            actions=["sns:Publish"],
                            resources=[
                                sns.topic_arn,
                            ],
                        )

                        lambda_.add_to_role_policy(sns_policy)

                        lambda_.add_environment(
                            k.resource_id + "_topic_arn",
                            getattr(self, k.resource_id).topic_arn or "",
                        )

                elif ResourceTypeGuard.is_websocket_permission(k):
                    if k.allow_publish:
                        ws: WebSocketApi = getattr(self, k.resource_id)

                        # Create a policy statement that allows invoking the @connections POST endpoint
                        policy_statement = aws_iam.PolicyStatement(
                            actions=["execute-api:ManageConnections"],
                            resources=[
                                f"arn:aws:execute-api:*:*:{ws.api_id}/*/@connections/*"
                            ],
                        )

                        # Attach the policy statement to the Lambda function's execution role
                        lambda_.add_to_role_policy(policy_statement)

                        # put the webocket api @connections url in the environment

                        lambda_.add_environment(
                            k.resource_id + "_api_id",
                            ws.api_id,
                        )

                elif ResourceTypeGuard.is_s3_permission(k):
                    bucket: aws_s3.Bucket = getattr(self, k.resource_id)

                    if any([k.allow_read, k.allow_write, k.allow_delete]):
                        lambda_.add_environment(
                            k.resource_id + "_bucket_name",
                            bucket.bucket_name or "",
                        )

                    if k.allow_read:
                        bucket.grant_read(lambda_)

                    if k.allow_write:
                        bucket.grant_write(lambda_)

                    if k.allow_delete:
                        bucket.grant_delete(lambda_)

                elif ResourceTypeGuard.is_sqs_permission(k):
                    queue: aws_sqs.Queue = getattr(self, k.resource_id)

                    if k.allow_publish:
                        queue.grant_send_messages(lambda_)

                        lambda_.add_environment(
                            k.resource_id + "_queue_url",
                            queue.queue_url or "",
                        )

    def add_websocket_gateway(self, resource: WsGateway):
        gw = add_ws_gateway_to_stack(
            stack=self,
            id=resource.id,
            stage_name=resource.stage_name,
            auto_deploy=resource.auto_deploy,
            websocket_name=resource.websocket_name,
        )

        if len(resource.permissions) > 1:
            raise ValueError("Websocket gateway can only invoke one lambda.")

        if resource.permissions:
            permission = resource.permissions[0]
            if (
                ResourceTypeGuard.is_lambda_permission(permission)
                and permission.allow_invoke
            ):
                func: aws_lambda.Function = getattr(self, permission.resource_id)

                for route in resource.routes:
                    gw.add_route(
                        route,
                        integration=WebSocketLambdaIntegration(
                            route + "_integration", func
                        ),
                    )

                # Allow the invoked functions to have access to @connections
                # Create a policy statement that allows invoking the @connections POST endpoint
                policy_statement = aws_iam.PolicyStatement(
                    actions=["execute-api:ManageConnections"],
                    resources=[f"arn:aws:execute-api:*:*:{gw.api_id}/*/@connections/*"],
                )

                # Attach the policy statement to the Lambda function's execution role
                func.add_to_role_policy(policy_statement)

                # Add stage name and api_id to the environment
                func.add_environment(
                    resource.id + "_api_id",
                    gw.api_id,
                )
                func.add_environment(resource.id + "_stage_name", resource.stage_name)

    def add_rest_api_gateway(self, resource: ApiGateway):
        if resource.existing_resource:
            gw = add_existing_rest_api_gateway_to_stack(
                stack=self,
                id=resource.id,
                rest_api_id=resource.rest_api_id or "",
                root_resource_id=resource.root_resource_id or "",
            )
        else:
            gw = add_rest_api_gateway_to_stack(
                stack=self,
                id=resource.id,
                rest_api_name=resource.rest_api_name,
                description=resource.description,
                endpoint_type=resource.endpoint_type,
                deploy=resource.deploy,
                stage_name=resource.stage_name,
            )

        if resource.permissions:
            for k in resource.permissions:
                if ResourceTypeGuard.is_lambda_permission(k) and k.allow_invoke:
                    func: aws_lambda.Function = getattr(self, k.resource_id)
                    gw.root.add_proxy(
                        any_method=True,
                        default_integration=aws_apigateway.LambdaIntegration(func),
                    )

                    func.grant_invoke(
                        aws_iam.ServicePrincipal("apigateway.amazonaws.com")
                    )

                    # We want to add the stage name as an environment variable
                    # For some api may require it but with the / at the front
                    func.add_environment(
                        LookupKey.api_path_prefix, "/" + resource.stage_name
                    )

    def add_dynamodb(self, resource: DynamoDB):
        add_dynamodb_to_stack(
            stack=self,
            id=resource.id,
            partition_key=resource.partition_key,
            sort_key=resource.sort_key,
            stream=resource.stream,
            billing_mode=resource.billing_mode,
            remove_on_delete=resource.remove_on_delete,
            local_secondary_indexes=resource.local_secondary_indexes,
            global_secondary_indexes=resource.global_secondary_indexes,
        )

    def add_s3_bucket(self, resource: S3Bucket):
        bucket = add_s3_bucket_to_stack(
            stack=self,
            id=resource.id,
            bucket_name=resource.bucket_name,
            remove_on_delete=resource.remove_on_delete,
        )

        if resource.permissions:
            for k in resource.permissions:
                if ResourceTypeGuard.is_lambda_permission(k) and k.allow_invoke:
                    func: aws_lambda.Function = getattr(self, k.resource_id)

                    bucket.add_object_created_notification(
                        aws_s3_notifications.LambdaDestination(func)
                    )

                    bucket.add_object_removed_notification(
                        aws_s3_notifications.LambdaDestination(func)
                    )

    def add_resource_to_stack(self, resource: AllResources):
        if isinstance(resource, LambdaFunction):
            self.add_lambda_function(resource)
        elif isinstance(resource, ApiGateway):
            self.add_rest_api_gateway(resource)
        elif isinstance(resource, WsGateway):
            self.add_websocket_gateway(resource)
        elif isinstance(resource, DynamoDB):
            self.add_dynamodb(resource)
        elif isinstance(resource, SnsTopic):
            self.add_sns_topic(resource)
        elif isinstance(resource, S3Bucket):
            self.add_s3_bucket(resource)
        elif isinstance(resource, RateSchedule):
            self.add_rate_schedule(resource)
        elif isinstance(resource, CronSchedule):
            self.add_cron_schedule(resource)
        elif isinstance(resource, SqsQueue):
            self.add_sqs_queue(resource)


class Infrastructure:
    def __init__(self, *, infrastructure: InfrastructureT) -> None:
        self._infrastructure = infrastructure

    @property
    def settings(self) -> InfraSettings:
        return self._infrastructure.settings

    @property
    def resources(self) -> dict[str, AllResources]:
        return self._infrastructure.resources

    @classmethod
    def from_dict(cls, value: dict) -> "Infrastructure":
        settings = InfraSettings(**value["settings"])
        resources = value["resources"]
        parsed_resources = parse_dict_to_resources(resources)

        return cls(
            infrastructure=InfrastructureT(
                settings=settings, resources=parsed_resources
            )
        )

    def get_stack(self) -> Stack:
        # This will look environment var if not provided
        return Stack(
            aws_account_id=self.settings.aws_account_id or default_aws_account_id(),
            aws_region=self.settings.aws_region or default_aws_region(),
            stack_name=self.settings.stack_name,
            environment_vars=self.settings.environment,
        )

    def stack(self) -> Stack:
        stack_builder = self.get_stack()
        ready: list[str] = []
        resources = self._infrastructure.resources

        # Build each resources based on dependencies resolution
        # Iterate until all resources are built
        while True:
            left = [
                resources[stack_resource].id
                for stack_resource in resources
                if resources[stack_resource].id not in ready
            ]

            for k in resources:
                resource = resources[k]
                permissions = resource.permissions

                # If resource is already built, skip
                if resource.id in ready:
                    continue

                # Check for permission dependency
                if permissions:
                    for i in permissions:
                        # Raise an error if the permission does not exist
                        # At all anywhere
                        if i.resource_id not in resources:
                            raise MissingStackResource(
                                f"""Resource {i.resource_id} not found in stack"""
                            )

                    # Check if all permission for this resource are resolved
                    if not all([i.resource_id in ready for i in permissions]):
                        continue

                # Add the resource to the base infrastructure
                stack_builder.add_resource_to_stack(resource)
                ready.append(resource.id)

            if not left:
                break

        return stack_builder



from .core import Fluxional, Extender
from .settings import Settings
from .tools import Environment, Event, Websocket
from .types import ApiEvent, LambdaContext, WsEvent, TaskEvent, StorageEvent

__all__ = [
    "Fluxional",
    "Environment",
    "Extender",
    "Settings",
    "ApiEvent",
    "LambdaContext",
    "WsEvent",
    "Event",
    "Websocket",
    "TaskEvent",
    "StorageEvent",
]



from dataclasses import dataclass, asdict, field
from .settings import Settings
from .infrastructure.resources import (
    ApiGateway,
    LambdaFunction,
    InfraResources,
    LambdaPermission,
    DynamoDB,
    DynamoDbPermission,
    WsGateway,
    S3Bucket,
    S3Permission,
    RateSchedule,
    CronSchedule,
    SqsQueue,
    SqsPermission,
    AllPermission,
)
from .infrastructure.types import (
    DynamoDBBillingModeT,
    DynamoDBKeyT,
    DynamoDBLsiT,
    DynamoDBStreamT,
    RateDurationUnitT,
    DynamoDBGsiT,
)
from typing import TypedDict


@dataclass(kw_only=True)
class Websocket:
    routes: list[str] = field(default_factory=list)
    websocket_lambda: LambdaFunction | None = None
    websocket_gateway: WsGateway | None = None

    def add_route(self, route: str):
        self.routes.append(route)

    def add_connect_route(self):
        self.add_route("$connect")

    def add_disconnect_route(self):
        self.add_route("$disconnect")

    def add_default_route(self):
        self.add_route("$default")


@dataclass(kw_only=True)
class Api:
    active: bool = False
    api_lambda: LambdaFunction | None = None
    api_gateway: ApiGateway | None = None


@dataclass(kw_only=True)
class Storage:
    active: bool = field(default=False)
    storage_lambda: LambdaFunction | None = None


class _ScheduleRateT(TypedDict):
    value: int | float
    unit: RateDurationUnitT
    schedule_name: str


class _ScheduleCronT(TypedDict):
    schedule_name: str
    day: str | None
    hour: str | None
    minute: str | None
    month: str | None
    week_day: str | None
    year: str | None


@dataclass(kw_only=True)
class Schedule:
    rate_schedule: list[_ScheduleRateT] = field(default_factory=list)
    cron_schedule: list[_ScheduleCronT] = field(default_factory=list)
    rate_schedule_lambda: LambdaFunction | None = None
    cron_schedule_lambda: LambdaFunction | None = None


@dataclass(kw_only=True)
class Event:
    active: bool = False
    event_lambda: LambdaFunction | None = None
    event_queue: SqsQueue | None = None


@dataclass(kw_only=True)
class App:
    websocket: Websocket = field(default_factory=Websocket)
    database: DynamoDB | None = None
    storage_bucket: S3Bucket | None = None
    settings: Settings
    ##
    api: Api = field(default_factory=Api)
    storage: Storage = field(default_factory=Storage)
    schedule: Schedule = field(default_factory=Schedule)
    event: Event = field(default_factory=Event)

    def _build_api(self, stack_name: str, resources: InfraResources):
        # Check
        if not self.api.active:
            return

        # Resources
        api_lambda = LambdaFunction(
            id=self.settings.system.default_api_lambda_id,
            function_name=f"{stack_name}_{self.settings.system.default_api_lambda_id}",
            existing_resource=False,
            **asdict(self.settings.build.api_lambda),
        )

        api_gateway = ApiGateway(
            id=self.settings.system.default_api_gateway_id,
            rest_api_name=f"{stack_name}_{self.settings.system.default_api_gateway_id}",
            **asdict(self.settings.build.api_gateway),
        )

        api_gateway.permissions.append(
            LambdaPermission(
                resource_id=api_lambda.id,
                resource_type=api_lambda.resource_type,
                allow_invoke=True,
            )
        )

        # Register
        resources[api_lambda.id] = api_lambda
        resources[api_gateway.id] = api_gateway
        self.api.api_lambda = api_lambda
        self.api.api_gateway = api_gateway

    def _build_db(self, stack_name: str, resources: InfraResources):
        if not self.database:
            return

        resources[self.database.id] = self.database

    def _build_websockets(self, stack_name: str, resources: InfraResources):
        # Check
        if not self.websocket.routes:
            return

        # Resources
        websocket_lambda = LambdaFunction(
            id=self.settings.system.default_websocket_lambda_id,
            function_name=f"{stack_name}_{self.settings.system.default_websocket_lambda_id}",
        )

        websocket_api_gateway = WsGateway(
            id=self.settings.system.default_websocket_gateway_id,
            routes=self.websocket.routes,
            websocket_name=f"{stack_name}_{self.settings.system.default_websocket_gateway_id}",
            stage_name=self.settings.build.websocket.stage_name,
            auto_deploy=self.settings.build.websocket.auto_deploy,
        )

        websocket_api_gateway.permissions.append(
            LambdaPermission(
                resource_id=websocket_lambda.id,
                resource_type=websocket_lambda.resource_type,
                allow_invoke=True,
            )
        )

        # Register
        resources[websocket_lambda.id] = websocket_lambda
        resources[websocket_api_gateway.id] = websocket_api_gateway
        self.websocket.websocket_gateway = websocket_api_gateway
        self.websocket.websocket_lambda = websocket_lambda

    def _build_storage(self, stack_name: str, resources: InfraResources):

        # @TODO: May need to refractor the whole storage workflow
        # Check
        if not self.settings.storage.enable and not self.storage.active:
            return

        # Resources
        self.add_storage_bucket(remove_on_delete=self.settings.storage.remove_on_delete)

        if isinstance(self.storage_bucket, S3Bucket):
            resources[self.storage_bucket.id] = self.storage_bucket

            storage_lambda = LambdaFunction(
                id=self.settings.system.default_storage_lambda_id,
                function_name=f"{stack_name}_{self.settings.system.default_storage_lambda_id}",
                existing_resource=False,
                **asdict(self.settings.build.storage_lambda),
            )

            self.storage_bucket.permissions.append(
                LambdaPermission(
                    resource_id=storage_lambda.id,
                    resource_type=storage_lambda.resource_type,
                    allow_invoke=True,
                )
            )

            # Register
            resources[storage_lambda.id] = storage_lambda
            self.storage.storage_lambda = storage_lambda

    def _build_rate_scheduled_tasks(self, stack_name: str, resources: InfraResources):

        # Check
        if not self.schedule.rate_schedule:
            return

        # Resources
        rate_schedules: list[RateSchedule] = []

        for sch in self.schedule.rate_schedule:
            schedule_name = sch["schedule_name"]
            value = sch["value"]
            unit = sch["unit"]

            rate_schedules.append(
                RateSchedule(
                    id=f"{stack_name}_{schedule_name}",
                    schedule_name=schedule_name,
                    value=value,
                    unit=unit,
                )
            )

        lambda_ = LambdaFunction(
            id=self.settings.system.default_rate_schedule_task_lambda_id,
            function_name=f"{stack_name}_{self.settings.system.default_rate_schedule_task_lambda_id}",
            existing_resource=False,
            **asdict(self.settings.build.scheduled_task_lambda),
        )

        for schedule in rate_schedules:
            schedule.permissions.append(
                LambdaPermission(
                    resource_id=lambda_.id,
                    resource_type=lambda_.resource_type,
                    allow_invoke=True,
                )
            )

            # Register
            resources[schedule.id] = schedule

        # Register
        resources[lambda_.id] = lambda_
        self.schedule.rate_schedule_lambda = lambda_

    def _build_cron_scheduled_tasks(self, stack_name: str, resources: InfraResources):
        # Check
        if not self.schedule.cron_schedule:
            return

        # Resources
        cron_schedules: list[CronSchedule] = []

        for sch in self.schedule.cron_schedule:
            schedule_name = sch["schedule_name"]
            day = sch["day"]
            hour = sch["hour"]
            minute = sch["minute"]
            month = sch["month"]
            week_day = sch["week_day"]
            year = sch["year"]

            cron_schedules.append(
                CronSchedule(
                    id=f"{stack_name}_{schedule_name}",
                    schedule_name=schedule_name,
                    day=day,
                    hour=hour,
                    minute=minute,
                    month=month,
                    week_day=week_day,
                    year=year,
                )
            )

        lambda_ = LambdaFunction(
            id=self.settings.system.default_cron_schedule_task_lambda_id,
            function_name=f"{stack_name}_{self.settings.system.default_cron_schedule_task_lambda_id}",
            existing_resource=False,
            **asdict(self.settings.build.scheduled_task_lambda),
        )

        for schedule in cron_schedules:
            schedule.permissions.append(
                LambdaPermission(
                    resource_id=lambda_.id,
                    resource_type=lambda_.resource_type,
                    allow_invoke=True,
                )
            )

            # Register
            resources[schedule.id] = schedule

        # Register
        resources[lambda_.id] = lambda_
        self.schedule.cron_schedule_lambda = lambda_

    def _build_events(self, stack_name: str, resources: InfraResources):
        # Check
        if not self.event.active:
            return

        # Resources
        event_lambda = LambdaFunction(
            id=self.settings.system.default_event_lambda_id,
            function_name=f"{stack_name}_{self.settings.system.default_event_lambda_id}",
            existing_resource=False,
            **asdict(self.settings.build.event_lambda),
        )

        queue = SqsQueue(
            id=self.settings.system.default_event_queue_id,
            queue_name=f"{stack_name}_{self.settings.system.default_event_queue_id}",
            # Visibility Timeout cannot be less than the lambda timeout
            visibility_timeout=self.settings.build.event_lambda.timeout,
        )

        queue.permissions.append(
            LambdaPermission(
                resource_id=event_lambda.id,
                resource_type=event_lambda.resource_type,
                allow_invoke=True,
            )
        )

        # Register
        resources[event_lambda.id] = event_lambda
        resources[queue.id] = queue
        self.event.event_lambda = event_lambda
        self.event.event_queue = queue

    @staticmethod
    def _append_permission_if_allowed(
        entity: LambdaFunction | None,
        permission: AllPermission,
        attributes: dict[str, bool],
    ):
        if entity and any(attributes.values()):
            for k, v in attributes.items():
                setattr(permission, k, v)
            entity.permissions.insert(0, permission)

    def _build_db_permissions(self):
        ## DYNAMODB PERMISSIONS ##
        if self.database and isinstance(self.database, DynamoDB):
            permission = DynamoDbPermission(
                resource_id=self.database.id,
                resource_type=self.database.resource_type,
                allow_read=False,
                allow_write=False,
            )

            # -> API
            self._append_permission_if_allowed(
                self.api.api_lambda,
                permission,
                {
                    "allow_read": self.settings.permissions.allow_api_to_read_from_db,
                    "allow_write": self.settings.permissions.allow_api_to_write_to_db,
                },
            )

            # -> WEBSOCKETS
            self._append_permission_if_allowed(
                self.websocket.websocket_lambda,
                permission,
                {
                    "allow_read": self.settings.permissions.allow_websockets_to_read_from_db,
                    "allow_write": self.settings.permissions.allow_websockets_to_write_to_db,
                },
            )
            # -> STORAGE
            self._append_permission_if_allowed(
                self.storage.storage_lambda,
                permission,
                {
                    "allow_read": self.settings.permissions.allow_storage_to_read_from_db,
                    "allow_write": self.settings.permissions.allow_storage_to_write_to_db,
                },
            )

            # -> ASYNC EVENTS
            self._append_permission_if_allowed(
                self.event.event_lambda,
                permission,
                {
                    "allow_read": self.settings.permissions.allow_events_to_read_from_db,
                    "allow_write": self.settings.permissions.allow_events_to_write_to_db,
                },
            )

            # -> TASKS
            self._append_permission_if_allowed(
                self.schedule.rate_schedule_lambda,
                permission,
                {
                    "allow_read": self.settings.permissions.allow_tasks_to_read_from_db,
                    "allow_write": self.settings.permissions.allow_tasks_to_write_to_db,
                },
            )

            self._append_permission_if_allowed(
                self.schedule.cron_schedule_lambda,
                permission,
                {
                    "allow_read": self.settings.permissions.allow_tasks_to_read_from_db,
                    "allow_write": self.settings.permissions.allow_tasks_to_write_to_db,
                },
            )

    def _build_storage_permissions(self):
        if not self.storage_bucket:
            return

        permission = S3Permission(
            resource_id=self.storage_bucket.id,
            resource_type=self.storage_bucket.resource_type,
            allow_read=False,
            allow_write=False,
            allow_delete=False,
        )

        # -> API
        self._append_permission_if_allowed(
            self.api.api_lambda,
            permission,
            {
                "allow_read": self.settings.permissions.allow_api_to_read_from_storage,
                "allow_write": self.settings.permissions.allow_api_to_write_to_storage,
                "allow_delete": self.settings.permissions.allow_api_to_delete_from_storage,
            },
        )

        # -> WEBSOCKETS
        self._append_permission_if_allowed(
            self.websocket.websocket_lambda,
            permission,
            {
                "allow_read": self.settings.permissions.allow_websockets_to_read_from_storage,
                "allow_write": self.settings.permissions.allow_websockets_to_write_to_storage,
                "allow_delete": self.settings.permissions.allow_websockets_to_delete_from_storage,
            },
        )

        # -> STORAGE
        self._append_permission_if_allowed(
            self.storage.storage_lambda,
            permission,
            {
                "allow_read": self.settings.permissions.allow_storage_to_read_from_storage,
                "allow_write": self.settings.permissions.allow_storage_to_write_to_storage,
                "allow_delete": self.settings.permissions.allow_storage_to_delete_from_storage,
            },
        )
        # -> EVENTS
        self._append_permission_if_allowed(
            self.event.event_lambda,
            permission,
            {
                "allow_read": self.settings.permissions.allow_events_to_read_from_storage,
                "allow_write": self.settings.permissions.allow_events_to_write_to_storage,
                "allow_delete": self.settings.permissions.allow_events_to_delete_from_storage,
            },
        )

        # -> TASKS
        self._append_permission_if_allowed(
            self.schedule.rate_schedule_lambda,
            permission,
            {
                "allow_read": self.settings.permissions.allow_tasks_to_read_from_storage,
                "allow_write": self.settings.permissions.allow_tasks_to_write_to_storage,
                "allow_delete": self.settings.permissions.allow_tasks_to_delete_from_storage,
            },
        )

        self._append_permission_if_allowed(
            self.schedule.cron_schedule_lambda,
            permission,
            {
                "allow_read": self.settings.permissions.allow_tasks_to_read_from_storage,
                "allow_write": self.settings.permissions.allow_tasks_to_write_to_storage,
                "allow_delete": self.settings.permissions.allow_tasks_to_delete_from_storage,
            },
        )

    def _build_events_permissions(self):
        if not self.event.event_queue:
            return

        permission = SqsPermission(
            resource_id=self.event.event_queue.id,
            resource_type=self.event.event_queue.resource_type,
            allow_publish=True,
        )

        # -> API
        self._append_permission_if_allowed(
            self.api.api_lambda,
            permission,
            {"allow_publish": self.settings.permissions.allow_api_to_run_events},
        )

        # -> WEBSOCKETS
        self._append_permission_if_allowed(
            self.websocket.websocket_lambda,
            permission,
            {"allow_publish": self.settings.permissions.allow_websockets_to_run_events},
        )

        # -> STORAGE
        self._append_permission_if_allowed(
            self.storage.storage_lambda,
            permission,
            {"allow_publish": self.settings.permissions.allow_storage_to_run_events},
        )

        # -> EVENTS
        self._append_permission_if_allowed(
            self.event.event_lambda,
            permission,
            {"allow_publish": self.settings.permissions.allow_events_to_run_events},
        )

        # -> TASKS
        self._append_permission_if_allowed(
            self.schedule.rate_schedule_lambda,
            permission,
            {"allow_publish": self.settings.permissions.allow_tasks_to_run_events},
        )

        self._append_permission_if_allowed(
            self.schedule.cron_schedule_lambda,
            permission,
            {"allow_publish": self.settings.permissions.allow_tasks_to_run_events},
        )

    def build_resources(
        self, as_dict: bool = False
    ) -> InfraResources | dict[str, dict]:
        """
        Build the infrastructure resources
        """
        resources: InfraResources = {}
        stack_name = self.settings.stack_name.lower()

        procedures = [
            ## Database ##
            self._build_db,
            ## API ##
            self._build_api,
            ## Websocket ##
            self._build_websockets,
            ## S3 ##
            self._build_storage,
            ## Scheduled Tasks ##
            self._build_rate_scheduled_tasks,
            self._build_cron_scheduled_tasks,
            ## Events ##
            self._build_events,
        ]

        permissions = [
            self._build_db_permissions,
            self._build_storage_permissions,
            self._build_events_permissions,
        ]

        for procedure in procedures:
            procedure(stack_name, resources)

        for permission in permissions:
            permission()

        if as_dict:
            return {k: asdict(v) for k, v in resources.items()}

        return resources

    def add_dynamodb(
        self,
        partition_key: DynamoDBKeyT | None = None,
        sort_key: DynamoDBKeyT | None = None,
        remove_on_delete: bool = True,
        local_secondary_indexes: list[DynamoDBLsiT] | None = None,
        global_secondary_indexes: list[DynamoDBGsiT] | None = None,
        # @NOTE Not currently supported
        stream: DynamoDBStreamT | None = None,
        billing_mode: DynamoDBBillingModeT | None = None,
    ) -> None:
        """
        Add a DynamoDB table to the application
        """

        dynamodb = DynamoDB(
            id=self.settings.system.default_dynamodb_id,
            partition_key=partition_key
            or self.settings.build.dynamodb.default_primary_key,
            sort_key=sort_key or self.settings.build.dynamodb.default_secondary_key,
            remove_on_delete=remove_on_delete,
            stream=stream or self.settings.build.dynamodb.default_stream,
            billing_mode=billing_mode
            or self.settings.build.dynamodb.default_billing_mode,
            local_secondary_indexes=local_secondary_indexes or [],
            global_secondary_indexes=global_secondary_indexes or [],
        )

        self.database = dynamodb

    def add_storage_bucket(
        self,
        bucket_name: str | None = None,
        remove_on_delete: bool = True,
    ) -> None:
        """
        Add an Storage bucket to the application
        """

        safe = lambda x: x.replace("_", "-").lower()  # noqa

        storage_bucket = S3Bucket(
            id=self.settings.system.default_storage_bucket_id,
            bucket_name=(
                bucket_name
                if bucket_name
                else safe(self.settings.stack_name)
                + safe(self.settings.system.default_storage_bucket_id)
            ),
            remove_on_delete=remove_on_delete,
        )

        self.storage_bucket = storage_bucket

    def set_api(self):
        self.api.active = True



import boto3  # type: ignore
import json
from typing import TypeVar, Generic, Optional
from .types import WsEvent
import os
from dataclasses import dataclass, field


@dataclass
class LookupKey:
    handler_context: str = "fluxional_handler_context"
    execution_context: str = "fluxional_execution_context"
    dynamodb_table_name: str = "fluxional_dynamodb_table_name"
    storage_bucket_name: str = "fluxional_storage_bucket_bucket_name"
    api_path_prefix: str = "fluxional_api_path_prefix"
    resource_id: str = "fluxional_resource_id"
    websocket_api_id: str = "fluxional_websocket_gateway_api_id"
    websocket_stage_name: str = "fluxional_websocket_gateway_stage_name"
    event_queue_url: str = "fluxional_event_queue_queue_url"


@dataclass
class Environment:
    # execution_context
    execution_context: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.execution_context)
    )

    # fluxional_run_context
    fluxional_handler_context: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.handler_context)
    )

    # If a dynamodb is in the stack this will have the table name
    dynamodb_table_name: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.dynamodb_table_name)
    )

    # If an storage bucket permission is in the stack this will have the bucket name
    storage_bucket_name: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.storage_bucket_name)
    )

    # This will be the stage name of the API Gateway but with a slash in front. ex: /prod
    api_path_prefix: str = field(
        default_factory=lambda: os.environ.get(LookupKey.api_path_prefix) or ""
    )

    # resource_id
    resource_id: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.resource_id)
    )

    # Web socket api id
    websocket_api_id: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.websocket_api_id)
    )

    # Web socket stage name
    websocket_stage_name: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.websocket_stage_name)
    )

    # Event sqs queue url
    event_queue_url: Optional[str] = field(
        default_factory=lambda: os.environ.get(LookupKey.event_queue_url)
    )

    # AWS Credentials
    aws_access_key_id: Optional[str] = field(
        default_factory=lambda: os.environ.get("AWS_ACCESS_KEY_ID")
    )
    aws_secret_access_key: Optional[str] = field(
        default_factory=lambda: os.environ.get("AWS_SECRET_ACCESS_KEY")
    )
    aws_session_token: Optional[str] = field(
        default_factory=lambda: os.environ.get("AWS_SESSION_TOKEN")
    )
    aws_region: Optional[str] = field(
        default_factory=lambda: os.environ.get("AWS_DEFAULT_REGION")
        or os.environ.get("AWS_REGION")
    )


T = TypeVar("T")


class Event(Generic[T]):
    def __init__(self):
        self._env = Environment()
        self._sqs = boto3.client("sqs", region_name=self._env.aws_region)

    def trigger(self, event_name: str, data: T) -> None:
        self._sqs.send_message(
            QueueUrl=self._env.event_queue_url,
            MessageBody=json.dumps(
                {
                    "event_name": event_name,
                    "data": data,
                }
            ),
        )


class Websocket:
    @staticmethod
    def post_to_connection(
        event: WsEvent, data: str, *, connection_id: str | None = None
    ):
        """
        Respond to a WebSocket connection id with data. if no connection_id is provided, it will be extracted from the event.
        """
        if not connection_id:
            connection_id = event["requestContext"]["connectionId"]

        api_gateway_management = boto3.client(
            "apigatewaymanagementapi",
            endpoint_url=f"https://{event['requestContext']['domainName']}/{event['requestContext']['stage']}",
        )
        api_gateway_management.post_to_connection(Data=data, ConnectionId=connection_id)



import os


def default_aws_account_id() -> str:
    AWS_ACCOUNT_ID = os.environ.get("AWS_ACCOUNT_ID")

    if not AWS_ACCOUNT_ID:
        raise ValueError("AWS_ACCOUNT_ID environment variable is required")

    return AWS_ACCOUNT_ID


def default_aws_region():
    AWS_REGION = os.environ.get("AWS_REGION")

    if not AWS_REGION:
        raise ValueError("AWS_REGION environment variable is required")

    return AWS_REGION


def default_aws_access_key_id() -> str:
    AWS_SECRET_KEY_ID = os.environ.get("AWS_ACCESS_KEY_ID")

    if not AWS_SECRET_KEY_ID:
        raise ValueError("AWS_SECRET_KEY_ID environment variable is required")

    return AWS_SECRET_KEY_ID


def default_aws_secret_access_key() -> str:
    AWS_SECRET_ACCESS_KEY = os.environ.get("AWS_SECRET_ACCESS_KEY")

    if not AWS_SECRET_ACCESS_KEY:
        raise ValueError("AWS_SECRET_ACCESS_KEY environment variable is required")

    return AWS_SECRET_ACCESS_KEY



from typing import TypedDict, Literal, Any, Optional, Dict, List
from typing_extensions import Protocol

_HttpMethods = Literal["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"]


## Api Gateway ##
class _RequestContext(TypedDict):
    resourcePath: str
    httpMethod: _HttpMethods
    path: str


class ApiEvent(TypedDict):
    resource: str
    path: str
    httpMethod: _HttpMethods
    requestContext: _RequestContext
    headers: dict
    multiValueHeaders: dict
    queryStringParameters: dict
    multiValueQueryStringParameters: dict
    pathParameters: dict
    stageVariables: dict
    isBase64Encoded: bool
    body: Any


## Lambda Context ##
# https://github.com/jordaneremieff/mangum/blob/main/mangum/types.pys
class LambdaCognitoIdentity(Protocol):
    """Information about the Amazon Cognito identity that authorized the request.

    **cognito_identity_id** - The authenticated Amazon Cognito identity.
    **cognito_identity_pool_id** - The Amazon Cognito identity pool that authorized the
    invocation.
    """

    cognito_identity_id: str
    cognito_identity_pool_id: str


class LambdaMobileClient(Protocol):
    """Mobile client information for the application and the device.

    **installation_id** - A unique identifier for an installation instance of an
    application.
    **app_title** - The title of the application. For example, "My App".
    **app_version_code** - The version of the application. For example, "V2.0".
    **app_version_name** - The version code for the application. For example, 3.
    **app_package_name** - The name of the package. For example, "com.example.my_app".
    """

    installation_id: str
    app_title: str
    app_version_name: str
    app_version_code: str
    app_package_name: str


class LambdaMobileClientContext(Protocol):
    """Information about client application and device when invoked via AWS Mobile SDK.

    **client** - A dict of name-value pairs that describe the mobile client application.
    **custom** - A dict of custom values set by the mobile client application.
    **env** - A dict of environment information provided by the AWS SDK.
    """

    client: LambdaMobileClient
    custom: Dict[str, Any]
    env: Dict[str, Any]


class LambdaContext(Protocol):
    """The context object passed to the handler function.

    **function_name** - The name of the Lambda function.
    **function_version** - The version of the function.
    **invoked_function_arn** - The Amazon Resource Name (ARN) that's used to invoke the
    function. Indicates if the invoker specified a version number or alias.
    **memory_limit_in_mb** - The amount of memory that's allocated for the function.
    **aws_request_id** - The identifier of the invocation request.
    **log_group_name** - The log group for the function.
    **log_stream_name** - The log stream for the function instance.
    **identity** - (mobile apps) Information about the Amazon Cognito identity that
    authorized the request.
    **client_context** - (mobile apps) Client context that's provided to Lambda by the
    client application.
    """

    function_name: str
    function_version: str
    invoked_function_arn: str
    memory_limit_in_mb: int
    aws_request_id: str
    log_group_name: str
    log_stream_name: str
    identity: Optional[LambdaCognitoIdentity]
    client_context: Optional[LambdaMobileClientContext]

    def get_remaining_time_in_millis(self) -> int:
        """Returns the number of milliseconds left before the execution times out."""
        ...  # pragma: no cover


## Websockets ##
class _WSRequestContext(TypedDict):
    routeKey: str
    eventType: str
    extendedRequestId: str
    requestTime: str
    messageDirection: str
    stage: str
    connectedAt: int
    requestTimeEpoch: int
    identity: dict
    requestId: str
    domainName: str
    connectionId: str
    apiId: str


class WsEvent(TypedDict, total=False):
    headers: dict
    multiValueHeaders: dict
    requestContext: _WSRequestContext
    body: str
    isBase64Encoded: bool


## Tasks ##
class TaskEvent(TypedDict):
    schedule_type: Literal["RateSchedule", "CronSchedule"]
    schedule_name: str


## S3 ##
class _StorageUserIdentity(TypedDict):
    principalId: str


class _StorageRequestParameters(TypedDict):
    sourceIPAddress: str


class _StorageResponseElements(TypedDict):
    x_amz_request_id: str
    x_amz_id_2: str


class _StorageOwnerIdentity(TypedDict):
    principalId: str


class _StorageBucket(TypedDict):
    name: str
    ownerIdentity: _StorageOwnerIdentity
    arn: str


class _StorageObject(TypedDict):
    key: str
    size: int
    eTag: str
    sequencer: str


class _StorageS3(TypedDict):
    s3SchemaVersion: str
    configurationId: str
    bucket: _StorageBucket
    object: _StorageObject


class _StorageRecord(TypedDict):
    eventVersion: str
    eventSource: str
    awsRegion: str
    eventTime: str
    eventName: str
    userIdentity: _StorageUserIdentity
    requestParameters: _StorageRequestParameters
    responseElements: _StorageResponseElements
    s3: _StorageS3


class StorageEvent(TypedDict):
    Records: List[_StorageRecord]



from typing import Literal, Optional, TypeGuard
from dataclasses import dataclass, field
from .types import (
    ServerlessResourcesT,
    DynamoDBKeyT,
    DynamoDBStreamT,
    DynamoDBBillingModeT,
    DynamoDBLsiT,
    RateDurationUnitT,
    DynamoDBGsiT,
)


@dataclass(kw_only=True)
class _CorePermission:
    resource_id: str
    resource_type: ServerlessResourcesT


@dataclass(kw_only=True)
class LambdaPermission(_CorePermission):
    allow_invoke: bool = field(default=False)
    permission_type: Literal["lambda_permission"] = field(default="lambda_permission")


@dataclass(kw_only=True)
class BaseDbPermission(_CorePermission):
    allow_read: bool
    allow_write: bool


@dataclass(kw_only=True)
class DynamoDbPermission(BaseDbPermission):
    permission_type: Literal["dynamodb_permission"] = field(
        default="dynamodb_permission"
    )


@dataclass(kw_only=True)
class SnsPermission(_CorePermission):
    allow_publish: bool
    permission_type: Literal["sns_permission"] = field(default="sns_permission")


@dataclass(kw_only=True)
class WebsocketPermission(_CorePermission):
    allow_publish: bool
    permission_type: Literal["websocket_permission"] = field(
        default="websocket_permission"
    )


@dataclass(kw_only=True)
class S3Permission(_CorePermission):
    allow_read: bool
    allow_write: bool
    allow_delete: bool
    permission_type: Literal["s3_permission"] = field(default="s3_permission")


@dataclass(kw_only=True)
class SqsPermission(_CorePermission):
    allow_publish: bool
    permission_type: Literal["sqs_permission"] = field(default="sqs_permission")


AllPermission = (
    LambdaPermission
    | DynamoDbPermission
    | SnsPermission
    | WebsocketPermission
    | S3Permission
    | SqsPermission
)


@dataclass(kw_only=True)
class _ResourceT:
    id: str
    # resource type
    resource_type: ServerlessResourcesT
    # Permissions are InfrastructureModels inject at build time
    permissions: list[AllPermission] = field(default_factory=list)
    # Resource exist
    existing_resource: bool = field(default=False)


@dataclass(kw_only=True)
class LambdaFunction(_ResourceT):
    function_name: str
    directory: str = field(default="./")
    dockerfile: str = field(default="Dockerfile")
    memory_size: int = field(default=128)
    timeout: int = field(default=30)
    description: str = field(default="")
    resource_type: Literal["lambda_function"] = field(default="lambda_function")


@dataclass(kw_only=True)
class ApiGateway(_ResourceT):
    rest_api_name: str
    description: str = field(default="fluxional_api_gateway")
    endpoint_type: Literal["regional", "edge"] = field(default="regional")
    deploy: bool = field(default=True)
    stage_name: str = field(default="prod")
    allowed_methods: list[str] = field(default_factory=list)
    allowed_origins: list[str] = field(default_factory=list)
    allowed_credentials: bool = field(default=False)
    allowed_headers: list[str] = field(default_factory=list)
    # Existing Resource
    rest_api_id: Optional[str] = field(default=None)
    root_resource_id: Optional[str] = field(default=None)
    resource_type: Literal["api_gateway"] = field(default="api_gateway")


@dataclass(kw_only=True)
class WsGateway(_ResourceT):
    websocket_name: str
    stage_name: str = field(default="prod")
    auto_deploy: bool = field(default=True)
    routes: list[str] = field(default_factory=list)
    resource_type: Literal["ws_gateway"] = field(default="ws_gateway")


@dataclass(kw_only=True)
class DynamoDB(_ResourceT):
    partition_key: DynamoDBKeyT
    sort_key: DynamoDBKeyT
    resource_type: Literal["dynamodb"] = field(default="dynamodb")
    stream: DynamoDBStreamT = field(default="new_image")
    billing_mode: DynamoDBBillingModeT = field(default="pay_per_request")
    remove_on_delete: bool = field(default=True)
    local_secondary_indexes: list[DynamoDBLsiT] = field(default_factory=list)
    global_secondary_indexes: list[DynamoDBGsiT] = field(default_factory=list)


@dataclass(kw_only=True)
class SnsTopic(_ResourceT):
    display_name: str
    resource_type: Literal["sns_topic"] = field(default="sns_topic")


@dataclass(kw_only=True)
class SqsQueue(_ResourceT):
    queue_name: str
    visibility_timeout: int = field(default=30)
    resource_type: Literal["sqs_queue"] = field(default="sqs_queue")


@dataclass(kw_only=True)
class S3Bucket(_ResourceT):
    bucket_name: str
    remove_on_delete: bool = field(default=True)
    resource_type: Literal["s3_bucket"] = field(default="s3_bucket")


@dataclass(kw_only=True)
class RateSchedule(_ResourceT):
    schedule_name: str
    value: int | float
    unit: RateDurationUnitT
    resource_type: Literal["rate_schedule"] = field(default="rate_schedule")


@dataclass(kw_only=True)
class CronSchedule(_ResourceT):
    schedule_name: str
    day: str | None = field(default=None)
    hour: str | None = field(default=None)
    minute: str | None = field(default=None)
    month: str | None = field(default=None)
    week_day: str | None = field(default=None)
    year: str | None = field(default=None)
    resource_type: Literal["cron_schedule"] = field(default="cron_schedule")


AllResources = (
    LambdaFunction
    | ApiGateway
    | DynamoDB
    | SnsTopic
    | WsGateway
    | S3Bucket
    | RateSchedule
    | CronSchedule
    | SqsQueue
)


@dataclass(kw_only=True)
class InfraSettings:
    stack_name: str
    aws_account_id: Optional[str]
    aws_region: Optional[str]
    environment: dict[str, str] = field(default_factory=dict)


InfraResources = dict[str, AllResources]


@dataclass(kw_only=True)
class InfrastructureT:
    settings: InfraSettings
    resources: InfraResources


class ResourceTypeGuard:
    @staticmethod
    def is_lambda_permission(val: AllPermission) -> TypeGuard[LambdaPermission]:
        return val.permission_type == "lambda_permission"

    @staticmethod
    def is_dynamodb_permission(val: AllPermission) -> TypeGuard[DynamoDbPermission]:
        return val.permission_type == "dynamodb_permission"

    @staticmethod
    def is_sns_permission(val: AllPermission) -> TypeGuard[SnsPermission]:
        return val.permission_type == "sns_permission"

    @staticmethod
    def is_websocket_permission(val: AllPermission) -> TypeGuard[WebsocketPermission]:
        return val.permission_type == "websocket_permission"

    @staticmethod
    def is_s3_permission(val: AllPermission) -> TypeGuard[S3Permission]:
        return val.permission_type == "s3_permission"

    @staticmethod
    def is_sqs_permission(val: AllPermission) -> TypeGuard[SqsPermission]:
        return val.permission_type == "sqs_permission"


def parse_dict_to_resources(resources: dict[str, dict]) -> dict[str, AllResources]:
    parsed_resources: dict[str, AllResources] = {}

    for k in resources:
        # Parse the permissions
        permissions = resources[k].get("permissions")
        parsed_permissions: list[AllPermission] = []
        if permissions:
            for perm in permissions:
                if perm["resource_type"] == "lambda_function":
                    parsed_permissions.append(LambdaPermission(**perm))
                if perm["resource_type"] == "dynamodb":
                    parsed_permissions.append(DynamoDbPermission(**perm))
                if perm["resource_type"] == "sns_topic":
                    parsed_permissions.append(SnsPermission(**perm))
                if perm["resource_type"] == "ws_gateway":
                    parsed_permissions.append(WebsocketPermission(**perm))
                if perm["resource_type"] == "s3_bucket":
                    parsed_permissions.append(S3Permission(**perm))
                if perm["resource_type"] == "sqs_queue":
                    parsed_permissions.append(SqsPermission(**perm))

        resources[k].pop("permissions", None)
        # Parse the resource
        if resources[k]["resource_type"] == "lambda_function":
            parsed_resources[k] = LambdaFunction(
                permissions=parsed_permissions, **resources[k]
            )
        elif resources[k]["resource_type"] == "api_gateway":
            parsed_resources[k] = ApiGateway(
                permissions=parsed_permissions, **resources[k]
            )

        elif resources[k]["resource_type"] == "ws_gateway":
            parsed_resources[k] = WsGateway(
                permissions=parsed_permissions, **resources[k]
            )

        elif resources[k]["resource_type"] == "dynamodb":
            parsed_resources[k] = DynamoDB(
                permissions=parsed_permissions, **resources[k]
            )

        elif resources[k]["resource_type"] == "sns_topic":
            parsed_resources[k] = SnsTopic(
                permissions=parsed_permissions, **resources[k]
            )

        elif resources[k]["resource_type"] == "sqs_queue":
            parsed_resources[k] = SqsQueue(
                permissions=parsed_permissions, **resources[k]
            )

        elif resources[k]["resource_type"] == "s3_bucket":
            parsed_resources[k] = S3Bucket(
                permissions=parsed_permissions, **resources[k]
            )

        elif resources[k]["resource_type"] == "rate_schedule":
            parsed_resources[k] = RateSchedule(
                permissions=parsed_permissions, **resources[k]
            )

        elif resources[k]["resource_type"] == "cron_schedule":
            parsed_resources[k] = CronSchedule(
                permissions=parsed_permissions, **resources[k]
            )

    return parsed_resources



from aws_cdk import (
    Stack,
    aws_lambda,
    Duration,
    aws_apigateway,
    aws_dynamodb,
    RemovalPolicy,
    aws_sns,
    aws_s3,
    aws_events,
    aws_sqs,
    CfnOutput,
)
from typing import Literal
from .types import (
    DynamoDBAttributeTypeT,
    DynamoDBKeyT,
    DynamoDBLsiT,
    DynamoDBBillingModeT,
    DynamoDBStreamT,
    DynamoDBGsiT,
)
import aws_cdk.aws_apigatewayv2_alpha as aws_apigateway_v2
from .types import RateDurationUnitT


def add_rate_schedule_to_stack(
    *,
    stack: Stack,
    id: str,
    unit: RateDurationUnitT,
    value: int | float,
    schedule_name: str,
) -> aws_events.Rule:
    durations = {
        "days": Duration.days,
        "hours": Duration.hours,
        "minutes": Duration.minutes,
        "seconds": Duration.seconds,
        "milliseconds": Duration.millis,
    }

    rule = aws_events.Rule(
        stack, schedule_name, schedule=aws_events.Schedule.rate(durations[unit](value))
    )

    setattr(stack, id, rule)

    return rule


def add_cron_schedule_to_stack(
    *,
    stack: Stack,
    id: str,
    schedule_name: str,
    day: str | None,
    hour: str | None,
    minute: str | None,
    month: str | None,
    week_day: str | None,
    year: str | None,
) -> aws_events.Rule:
    rule = aws_events.Rule(
        stack,
        schedule_name,
        schedule=aws_events.Schedule.cron(
            minute=minute,
            hour=hour,
            day=day,
            month=month,
            year=year,
            week_day=week_day,
        ),
    )

    setattr(stack, id, rule)

    return rule


def add_lambda_function_to_stack(
    *,
    stack: Stack,
    id: str,
    function_name: str,
    directory: str,
    file: str,
    memory_size: int,
    timeout: int,
    description: str,
) -> aws_lambda.Function:
    ecr_image = aws_lambda.EcrImageCode.from_asset_image(directory=directory, file=file)

    lambda_function = aws_lambda.Function(
        stack,
        id=id,
        description=description,
        code=ecr_image,
        handler=aws_lambda.Handler.FROM_IMAGE,
        runtime=aws_lambda.Runtime.FROM_IMAGE,
        function_name=function_name,
        memory_size=memory_size,
        timeout=Duration.seconds(timeout),
        log_format=aws_lambda.LogFormat.JSON.value,
    )

    setattr(stack, id, lambda_function)

    return lambda_function


def add_ws_gateway_to_stack(
    *,
    stack: Stack,
    id: str,
    stage_name: str,
    websocket_name: str,
    auto_deploy: bool,
) -> aws_apigateway_v2.WebSocketApi:
    web_socket_api = aws_apigateway_v2.WebSocketApi(stack, id, api_name=websocket_name)
    aws_apigateway_v2.WebSocketStage(
        stack,
        id + "_stage",
        web_socket_api=web_socket_api,
        stage_name=stage_name,
        auto_deploy=auto_deploy,
    )

    # Output the websocket url
    CfnOutput(
        stack,
        "WebsocketUrl",
        value=f"{web_socket_api.api_endpoint}/{stage_name}",
        description="The URL of the websocket",
    )

    setattr(stack, id, web_socket_api)

    return web_socket_api


def add_rest_api_gateway_to_stack(
    *,
    stack: Stack,
    id: str,
    rest_api_name: str,
    description: str,
    stage_name: str,
    deploy: bool,
    endpoint_type: Literal["regional", "edge"],
) -> aws_apigateway.RestApi:
    """
    Represents an api gateway. Adds it to the passed stack.
    """

    endpoint_type_mapper = {
        "regional": aws_apigateway.EndpointType.REGIONAL,
        "edge": aws_apigateway.EndpointType.EDGE,
    }

    api = aws_apigateway.RestApi(
        stack,
        id,
        rest_api_name=rest_api_name,
        description=description,
        endpoint_types=[endpoint_type_mapper[endpoint_type]],
        deploy=deploy,
        deploy_options=aws_apigateway.StageOptions(stage_name=stage_name),
    )

    setattr(stack, id, api)

    return api


def add_existing_rest_api_gateway_to_stack(
    *, stack: Stack, id: str, rest_api_id: str, root_resource_id: str
) -> aws_apigateway.IRestApi:
    api = aws_apigateway.RestApi.from_rest_api_attributes(
        stack, id, rest_api_id=rest_api_id, root_resource_id=root_resource_id
    )

    setattr(stack, id, api)

    return api


def add_dynamodb_to_stack(
    *,
    stack: Stack,
    id: str,
    partition_key: DynamoDBKeyT,
    sort_key: DynamoDBKeyT,
    stream: DynamoDBStreamT,
    billing_mode: DynamoDBBillingModeT,
    remove_on_delete: bool,
    local_secondary_indexes: list[DynamoDBLsiT],
    global_secondary_indexes: list[DynamoDBGsiT],
):
    stream_mapper: dict[DynamoDBStreamT, aws_dynamodb.StreamViewType] = {
        "new_and_old_images": aws_dynamodb.StreamViewType.NEW_AND_OLD_IMAGES,
        "new_image": aws_dynamodb.StreamViewType.NEW_IMAGE,
        "old_image": aws_dynamodb.StreamViewType.OLD_IMAGE,
    }

    billing_mode_mapper: dict[DynamoDBBillingModeT, aws_dynamodb.BillingMode] = {
        "pay_per_request": aws_dynamodb.BillingMode.PAY_PER_REQUEST,
    }

    attribute_type_mapper: dict[DynamoDBAttributeTypeT, aws_dynamodb.AttributeType] = {
        "string": aws_dynamodb.AttributeType.STRING,
        "number": aws_dynamodb.AttributeType.NUMBER,
        "binary": aws_dynamodb.AttributeType.BINARY,
    }

    removal_policy = RemovalPolicy.DESTROY if remove_on_delete else RemovalPolicy.RETAIN

    db = aws_dynamodb.Table(
        stack,
        id,
        partition_key=aws_dynamodb.Attribute(
            name=partition_key["key_name"],
            type=attribute_type_mapper[partition_key["key_type"]],
        ),
        sort_key=aws_dynamodb.Attribute(
            name=sort_key["key_name"],
            type=attribute_type_mapper[sort_key["key_type"]],
        ),
        billing_mode=billing_mode_mapper[billing_mode],
        stream=stream_mapper[stream],
        removal_policy=removal_policy,
    )

    for lsi in local_secondary_indexes:
        db.add_local_secondary_index(
            sort_key=aws_dynamodb.Attribute(
                name=lsi["sort_key"]["key_name"],
                type=attribute_type_mapper[lsi["sort_key"]["key_type"]],
            ),
            index_name=lsi["index_name"],
        )

    for gsi in global_secondary_indexes:
        db.add_global_secondary_index(
            partition_key=aws_dynamodb.Attribute(
                name=gsi["partition_key"]["key_name"],
                type=attribute_type_mapper[gsi["partition_key"]["key_type"]],
            ),
            sort_key=aws_dynamodb.Attribute(
                name=gsi["sort_key"]["key_name"],
                type=attribute_type_mapper[gsi["sort_key"]["key_type"]],
            ),
            index_name=gsi["index_name"],
        )

    setattr(stack, id, db)

    return db


def add_sns_topic_to_stack(
    *, stack: Stack, id: str, display_name: str
) -> aws_sns.Topic:
    sns = aws_sns.Topic(
        stack,
        id,
        display_name=display_name,
    )

    setattr(stack, id, sns)

    return sns


def add_s3_bucket_to_stack(
    *, stack: Stack, id: str, bucket_name: str, remove_on_delete: bool
) -> aws_s3.Bucket:
    bucket = aws_s3.Bucket(
        stack,
        id,
        bucket_name=bucket_name,
        removal_policy=(
            RemovalPolicy.DESTROY if remove_on_delete else RemovalPolicy.RETAIN
        ),
        auto_delete_objects=remove_on_delete,
    )

    setattr(stack, id, bucket)

    return bucket


def add_sqs_queue_to_stack(
    *, stack: Stack, id: str, queue_name: str, visibility_timeout: int
) -> aws_sqs.Queue:
    queue = aws_sqs.Queue(
        stack,
        id,
        queue_name=queue_name,
        visibility_timeout=Duration.seconds(visibility_timeout),
    )

    setattr(stack, id, queue)

    return queue



class FailedImageBuild(Exception):
    """
    Exception raised when the docker image fails to build
    """


class FailedContainerBuild(Exception):
    """
    Exception raised when the docker container fails to build
    """


class FailedToRunContainer(Exception):
    """
    Exception raised when the docker container fails to run
    """



import typer
import os
import sys
from typing import Callable

from dotenv import load_dotenv

load_dotenv()


app = typer.Typer()
cwd = os.getcwd()


def import_method_from_handler(*, handler: str, path: str) -> Callable:
    """
    Import a method from a handler
    """
    sys.path.append(path)
    file, method = handler.split(".")
    test = __import__(file)
    return getattr(test, method)


@app.command()
def deploy(
    handler: str,
    path: str = cwd,
    show_logs: bool = True,
):
    func = import_method_from_handler(handler=handler, path=path)
    func(
        {
            "fluxional_event": "cli_deploy",
            "show_logs": show_logs,
        },
        {"handler": handler},
    )


@app.command()
def destroy(handler: str, path: str = cwd, show_logs: bool = True):
    func = import_method_from_handler(handler=handler, path=path)
    func(
        {
            "fluxional_event": "cli_destroy",
            "show_logs": show_logs,
        },
        {"handler": handler},
    )


@app.command()
def sync(handler: str, path: str = cwd):
    func = import_method_from_handler(handler=handler, path=path)
    func({"fluxional_event": "cli_sync"}, {"handler": handler})


@app.command()
def dev(handler: str, path: str = cwd):
    func = import_method_from_handler(handler=handler, path=path)
    func({"fluxional_event": "cli_dev"}, {"handler": handler})


def run_command_line():
    return app()



from typing import Any
from fluxional.types import HandlerFunctionT, AsyncHandlerFunctionT, AsyncTypeGuard
from fluxional.core.events import (
    is_http_event,
    is_fluxional_event,
    is_websocket_event,
    is_dev_context,
    is_s3_bucket_create_event,
    is_s3_bucket_delete_event,
    is_rate_schedule_event,
    is_cron_schedule_event,
    is_sqs_event,
    S3_ACTIONS,
)
from fluxional.core.app import App
from fluxional.exceptions import NoHandlerFound
from fluxional.utils import (
    default_aws_account_id,
    default_aws_region,
    default_aws_secret_access_key,
    default_aws_access_key_id,
)
from .settings import Settings
import asyncio
import sys
from uuid import uuid4
import time
import os
import json
from fluxional.core.tools import LookupKey
from .types import LambdaContext

_HandlerFunctionT = HandlerFunctionT | AsyncHandlerFunctionT


class Handlers:
    def __init__(
        self, *, settings: Settings | None = None, app: App | None = None
    ) -> None:
        self._settings = settings
        self._app = app
        self._fluxional_handlers: list[_HandlerFunctionT] = []
        self._http_handlers: list[_HandlerFunctionT] = []
        self._websocket_handlers: dict[str, _HandlerFunctionT] = {}
        self._storage_handlers: dict[str, _HandlerFunctionT] = {}
        self._rate_schedule_handlers: dict[str, _HandlerFunctionT] = {}
        self._cron_schedule_handlers: dict[str, _HandlerFunctionT] = {}
        self._sqs_handlers: dict[str, _HandlerFunctionT] = {}

    def _default_handlers(self) -> dict[str, HandlerFunctionT]:
        return {
            "deployment_handler": lambda event, context: deployment_handler(
                event, context, settings=self._settings
            ),
            "cli_dev_handler": lambda event, context: cli_dev_handler(
                event, context, settings=self._settings
            ),
        }

    def _register_default_handlers(self) -> None:
        handlers = self._default_handlers()

        for _, handler in handlers.items():
            self.add_fluxional_handler(handler)

    def add_fluxional_handler(self, handler: _HandlerFunctionT):
        self._fluxional_handlers.append(handler)

    def get_result(self, handler: _HandlerFunctionT, event, context):
        result: Any = None

        # We need to capture the exception here so that we can
        # know for unhandled errors
        exception: BaseException | None = None

        try:
            if AsyncTypeGuard.is_async(handler):
                loop = asyncio.get_event_loop()
                result = loop.run_until_complete(handler(event, context))

            else:
                result = handler(event, context)

        except Exception as e:
            exception = e

        if result is not None:
            return result

        if exception is not None:
            raise exception

    def sync_or_async(self, event, context) -> Any:

        handlers = []

        # Handler context
        if is_dev_context(event, context):
            handlers = [lambda e, c: dev_handler(e, c, settings=self._settings)]

        # Handle fluxional events
        elif is_fluxional_event(event, context):
            handlers = self._fluxional_handlers
            for handler in handlers:
                result = self.get_result(handler, event, context)
                if result:
                    return result

        # Handle http events
        elif is_http_event(event, context):
            handlers = self._http_handlers

        # Handle websocket events
        elif is_websocket_event(event, context):
            route = is_websocket_event(event, context)
            if route in self._websocket_handlers and isinstance(route, str):
                handlers = [self._websocket_handlers[route]]

        elif is_s3_bucket_create_event(event, context):
            if "create" in self._storage_handlers:
                handlers = [self._storage_handlers["create"]]

        elif is_s3_bucket_delete_event(event, context):
            if "delete" in self._storage_handlers:
                handlers = [self._storage_handlers["delete"]]

        elif is_rate_schedule_event(event, context):
            if event["schedule_name"] in self._rate_schedule_handlers:
                handlers = [self._rate_schedule_handlers[event["schedule_name"]]]

        elif is_cron_schedule_event(event, context):
            if event["schedule_name"] in self._cron_schedule_handlers:
                handlers = [self._cron_schedule_handlers[event["schedule_name"]]]

        elif is_sqs_event(event, context):
            # @TODO: Probably will need to be refractored
            body = json.loads(event["Records"][0]["body"])
            if body["event_name"] in self._sqs_handlers:
                handlers = [
                    lambda _, c: self._sqs_handlers[body["event_name"]](body["data"], c)
                ]

        if not handlers:
            raise NoHandlerFound("No handler found")

        # This may need to be refractored at some point
        handler = handlers[0]
        return self.get_result(handler, event, context)

    def synth_handler(self) -> Any:
        """Special case for when file is called directly with
        CLI - IE generating infrastructure"""

        if "--synth" in sys.argv:
            return True

    def handler(self) -> Any:
        self._register_default_handlers()
        return self.sync_or_async

    def add_api_handler(self, handler: _HandlerFunctionT) -> None:
        self._http_handlers.append(handler)

    def add_websocket_route_handler(
        self,
        route: str,
        handler: _HandlerFunctionT,
    ) -> None:
        self._websocket_handlers[route] = handler

    def add_storage_handler(
        self, action: S3_ACTIONS, handler: _HandlerFunctionT
    ) -> None:
        self._storage_handlers[action] = handler


def cli_dev_handler(event: dict, context: Any, settings: Settings | None = None):
    if not event.get("fluxional_event"):
        return None

    if event["fluxional_event"] != "cli_dev":
        return None

    if not settings:
        raise ValueError("Settings are required to run RTL.")

    from fluxional.dev import run_dev

    run_dev(
        settings.stack_name,
        handler=context["handler"],
    )

    return True


def deployment_handler(
    event: dict, context: Any, settings: Settings | None = None
) -> bool | None:
    if not event.get("fluxional_event"):
        return None

    if event["fluxional_event"] not in ["cli_deploy", "cli_destroy"]:
        return None

    if not settings:
        raise ValueError("Settings are required to deploy.")

    handler: str = context["handler"]
    show_logs: bool = event.get("show_logs", False)
    stack_name = settings.stack_name
    aws_account_id = settings.credentials.aws_account_id
    aws_region = settings.credentials.aws_region
    aws_access_key_id = settings.credentials.aws_access_key_id
    aws_secret_access_key = settings.credentials.aws_secret_access_key
    dependencies = settings.build.dependencies
    environment = settings.build.environment
    enable_otel = settings.monitoring.otel.enable
    enable_telemetry = settings.monitoring.telemetry.enable

    if settings.development.enable_local:
        # Turn on development mode
        environment[LookupKey.handler_context] = "development"

    if enable_otel:
        # Add vars necessary for otel
        environment["OTEL_EXPORTER_OTLP_ENDPOINT"] = (
            settings.monitoring.otel.exporter_otlp_endpoint
        )
        environment["OTEL_EXPORTER_OTLP_HEADERS"] = (
            settings.monitoring.otel.exporter_otlp_headers
        )

        environment["AWS_LAMBDA_EXEC_WRAPPER"] = settings.system.aws_lambda_exec_wrapper

        environment["OTEL_SERVICE_NAME"] = settings.monitoring.otel.service_name

    requirements_file = settings.build.requirements_file
    build_path = settings.build.build_path or "."
    container_command = settings.system.container_command

    from fluxional.deployment import CDKEngine

    engine = CDKEngine(
        aws_account_id=aws_account_id or default_aws_account_id(),
        aws_region=aws_region or default_aws_region(),
        aws_access_key_id=aws_access_key_id or default_aws_access_key_id(),
        aws_secret_access_key=aws_secret_access_key or default_aws_secret_access_key(),
        build_path=build_path,
        tag=f"{stack_name.lower()}_base_image",
    )

    if event["fluxional_event"] == "cli_destroy":
        engine.destroy(
            handler,
            dependencies=dependencies,
            environment=environment,
            requirements_file=requirements_file,
            command=container_command,
            lambda_handler=settings.system.lambda_handler,
            show_logs=show_logs,
        )

    else:
        engine.deploy(
            handler,
            dependencies=dependencies,
            environment=environment,
            requirements_file=requirements_file,
            command=container_command,
            lambda_handler=settings.system.lambda_handler,
            show_logs=show_logs,
            include_otel=enable_otel,
            include_fte=enable_telemetry,
        )

    return True


class _DevState:
    subscriber_id: str | None = None
    acknowledged: bool = False
    response: dict | None = None

    def to_dict(self, value: bytes) -> dict:
        return json.loads(value.decode("utf-8"))

    def acknowledge(self, **kwargs):
        if not self.acknowledged:
            self.acknowledged = True
            # Set the subscriber id
            self.subscriber_id = self.to_dict(kwargs["payload"])["subscriber_id"]

    def on_event(self, **kwargs):
        payload = self.to_dict(kwargs["payload"])
        if payload["subscriber_id"] == self.subscriber_id:
            self.response = payload["response"]


def dev_handler(
    event: dict,
    context: LambdaContext,
    settings: Settings | None = None,
    ack_timeout=15,
    timeout=15,
    state=_DevState,
):
    state = state()

    if not settings:
        raise ValueError("Settings are required to run RTL.")

    from fluxional.dev.client import DevClient

    event_id = str(uuid4())
    stack_name = settings.stack_name

    client = DevClient()
    client.connect()

    client.subscribe(f"fluxional/response/{event_id}", state.on_event)
    client.subscribe(f"fluxional/acknowledge/{event_id}", state.acknowledge)

    # execution context
    starts_with = ["AWS_", "fluxional"]
    execution_context = {
        key: value
        for key, value in dict(os.environ).items()
        if any(key.startswith(start) for start in starts_with)
    }

    pub = client.publish(
        f"fluxional/events/{stack_name}",
        {
            "type": "new_event",
            "event_id": event_id,
            "event": event,
            "execution_context": execution_context,
        },
    )

    pub.result(timeout=timeout)

    # This needs to be acknowledged by some subscriber within
    # the timeout
    start = time.time()
    while True:
        if time.time() - start > ack_timeout:
            return {"statusCode": 500, "body": "Failed to acknowledge event."}

        if state.acknowledged:
            break

    # Only http needs to wait for a response
    # The rest are async
    start = time.time()
    if is_http_event(event, context):
        # But less than timeout which is pretty long already
        while True:
            if time.time() - start > timeout:
                return {"statusCode": 500, "body": "Failed to get a response in time."}

            if state.response:
                return state.response
    else:
        return {"statusCode": 200}
